import numpy
from typing import NoReturn
from cryspy.A_functions_base.function_1_matrices import\
    calc_product_matrices, calc_product_matrix_vector
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.C_item_loop_classes.cl_1_cell import Cell


class ReflnSusceptibility(ItemN):
    """
    Data items in the REFLN_SUSCEPTIBILITY category record details about the
    susceptibility structure factor tensor and magnetization structure factor 
    tensor calculated for different reflections.

    Mandatory attributes:
        - ub_11, ub_12, ub_13, ub_21, ub_22, ub_23, ub_31, ub_32, ub_33

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """
    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = (
        "sintlambda", "d_spacing", "chi_11_calc", "chi_12_calc", "chi_13_calc", 
        "chi_21_calc", "chi_22_calc", "chi_23_calc", "chi_31_calc",
        "chi_32_calc", "chi_33_calc", "moment_11_calc", "moment_12_calc",
        "moment_13_calc", "moment_21_calc", "moment_22_calc", "moment_23_calc", 
        "moment_31_calc", "moment_32_calc", "moment_33_calc")
    ATTR_OPTIONAL_TYPES = (float, float, complex, complex, complex,
                           complex, complex, complex, complex, complex, 
                           complex, complex, complex, complex, complex,
                           complex, complex, complex, complex, complex)
    ATTR_OPTIONAL_CIF = (
        "sintlambda", "d_spacing", "chi_11_calc", "chi_12_calc", "chi_13_calc", 
        "chi_21_calc", "chi_22_calc", "chi_23_calc", "chi_31_calc",
        "chi_32_calc", "chi_33_calc", "moment_11_calc", "moment_12_calc",
        "moment_13_calc", "moment_21_calc", "moment_22_calc", "moment_23_calc", 
        "moment_31_calc", "moment_32_calc", "moment_33_calc")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"sintlambda": "{:.5f}", "d_spacing": "{:.5f}",
                 "chi_11_calc": "{:.2f}", "chi_12_calc": "{:.2f}",
                 "chi_13_calc": "{:.2f}", "chi_21_calc": "{:.2f}",
                 "chi_22_calc": "{:.2f}", "chi_23_calc": "{:.2f}",
                 "chi_31_calc": "{:.2f}", "chi_32_calc": "{:.2f}",
                 "chi_33_calc": "{:.2f}", "moment_11_calc": "{:.2f}",
                 "moment_12_calc": "{:.2f}", "moment_13_calc": "{:.2f}",
                 "moment_21_calc": "{:.2f}", "moment_22_calc": "{:.2f}",
                 "moment_23_calc": "{:.2f}", "moment_31_calc": "{:.2f}",
                 "moment_32_calc": "{:.2f}", "moment_33_calc": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "refln_susceptibility"

    def __init__(self, **kwargs) -> NoReturn:
        super(ReflnSusceptibility, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"sintlambda": 0, "d_spacing": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class ReflnSusceptibilityL(LoopN):
    """
    Description of AtomSite in loop.

    """
    ITEM_CLASS = ReflnSusceptibility
    ATTR_INDEX = "id"
    def __init__(self, loop_name = None) -> NoReturn:
        super(ReflnSusceptibilityL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#  loop_
#  _refln_susceptibility.index_h    
#  _refln_susceptibility.index_k    
#  _refln_susceptibility.index_l    
#  _refln_susceptibility.d_spacing       
#  _refln_susceptibility.sintlambda
#  _refln_susceptibility.chi_11_calc     
#  _refln_susceptibility.chi_12_calc     
#  _refln_susceptibility.chi_13_calc     
#  _refln_susceptibility.chi_21_calc     
#  _refln_susceptibility.chi_22_calc     
#  _refln_susceptibility.chi_23_calc     
#  _refln_susceptibility.chi_31_calc     
#  _refln_susceptibility.chi_32_calc     
#  _refln_susceptibility.chi_33_calc     
#  _refln_susceptibility.moment_11_calc  
#  _refln_susceptibility.moment_12_calc  
#  _refln_susceptibility.moment_13_calc  
#  _refln_susceptibility.moment_21_calc  
#  _refln_susceptibility.moment_22_calc  
#  _refln_susceptibility.moment_23_calc  
#  _refln_susceptibility.moment_31_calc  
#  _refln_susceptibility.moment_32_calc  
#  _refln_susceptibility.moment_33_calc  
#  2 0 0 4.52 0.123 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j  0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j
#  0 2 0 4.52 0.123 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j  0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j 0+0j
#   """

# obj = ReflnSusceptibilityL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.sintlambda, end="\n\n")
# print(obj.numpy_chi_33_calc, end="\n\n")



"""Description of MEM class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_2_mem import calc_moment_perp, \
    calc_fm_by_density

from cryspy.B_parent_classes.cl_4_global import GlobalN


from cryspy.C_item_loop_classes.cl_1_mem_parameters import MEMParameters
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs

from cryspy.C_item_loop_classes.cl_2_section import SectionL

from cryspy.C_item_loop_classes.cl_3_density_point import DensityPointL

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn

from cryspy.F_functions_data.script_1_mem import maximize_entropy, \
    refine_susceptibility, make_cycle, calc_moments_in_unit_cell, \
    calc_section_for_mem


class MEM(GlobalN):
    """
    Class to describe RhoChi class.

    Attributes
    ----------
        - crystal_#name (mandatory)
        - diffrn_#name
        - density_point_#name

    Methods
    -------
        - crystals()
        - experiments()
        - maximize_entropy()
        - refine_susceptibility()
        - make_cycle()
        - apply_constraints()
    """

    CLASSES_MANDATORY = (Crystal, Diffrn)
    CLASSES_OPTIONAL = (DensityPointL, MEMParameters, SectionL)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "mem"

    # default values for the parameters
    D_DEFAULT = {"mem_parameters": MEMParameters()}

    def __init__(self, global_name=None, **kwargs) -> NoReturn:
        super(MEM, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["global_name"] = global_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self):
        """Form object."""
        self.apply_constraint()

    def apply_constraint(self):
        """Apply constraints."""
        for item in self.items:
            if isinstance(item, Crystal):
                item.apply_constraint()

    def experiments(self):
        """List of expreiments."""
        return [item for item in self.items if isinstance(item, Diffrn)]

    def crystals(self):
        """List of crystals."""
        return [item for item in self.items if isinstance(item, Crystal)]

    def save_to_file_den(self, f_name: str = "file.den", label_atom=None,
                         f_background: str = "file_back.den"):
        """Save density to ".den" files."""
        crystal = self.crystals()[0]
        space_group = crystal.space_group
        cell = crystal.cell
        mem_parameters = self.mem_parameters
        flag_two_channel = mem_parameters.method == "2channel"
        if flag_two_channel:
            l_label_atom = None
        else:
            atom_site_susceptibility = crystal.atom_site_susceptibility
            l_label_atom = atom_site_susceptibility.label
        density_point = self.density_point
        density_point.save_to_file_den(
            mem_parameters, space_group, cell, f_name=f_name,
            l_label_atom=l_label_atom, f_background=f_background)

    def create_prior_density(self):
        """Create prior denisity."""
        crystal = self.crystals()[0]  # FIXME:
        atom_site_susceptibility = crystal.atom_site_susceptibility
        space_group = crystal.space_group
        space_group_symop = space_group.full_space_group_symop
        cell = crystal.cell
        atom_site = crystal.atom_site
        mem_parameters = self.mem_parameters
        prior_density = mem_parameters.prior_density
        points_a = mem_parameters.points_a
        points_b = mem_parameters.points_b
        points_c = mem_parameters.points_c
        flag_two_channel = mem_parameters.method == "2channel"

        l_magnetic_labes = atom_site_susceptibility.label
        density_point = DensityPointL()
        if prior_density == "core":
            atom_electron_configuration = crystal.atom_electron_configuration
            density_point.create_core_density(
                space_group_symop, cell, atom_site,
                atom_electron_configuration, points_a=points_a,
                points_b=points_b, points_c=points_c,
                flag_two_channel=flag_two_channel)
        else:
            density_point.create_flat_density(
                space_group_symop, cell, atom_site,
                l_magnetic_labes=l_magnetic_labes, points_a=points_a,
                points_b=points_b, points_c=points_c,
                flag_two_channel=flag_two_channel)

        self.add_items([density_point])

    def calc_fr(self):
        """Calculate Flip Ratios for diffraction experiments."""
        crystal = self.crystals()[0]  # FIXME:
        l_diffrn = self.experiments()  # FIXME:
        density_point = self.density_point
        mem_parameters = self.mem_parameters
        chi_iso_ferro = mem_parameters.chi_ferro
        chi_iso_antiferro = mem_parameters.chi_antiferro

        flag_two_channel = mem_parameters.method == "2channel"

        cell = crystal.cell
        space_group = crystal.space_group
        space_group_symop = space_group.full_space_group_symop
        atom_site_susceptibility = crystal.atom_site_susceptibility

        # FIXME: temporary solution of calculation rbs_i if it's not defined
        density_point.calc_rbs_i(
            space_group_symop, points_a=mem_parameters.points_a,
            points_b=mem_parameters.points_b, points_c=mem_parameters.points_c)

        total_peaks = 0

        den_i = numpy.array(density_point.density, dtype=float)
        den_ferro_i = numpy.array(density_point.density_ferro, dtype=float)
        den_antiferro_i = numpy.array(density_point.density_antiferro,
                                      dtype=float)
        mult_i = numpy.array(density_point.multiplicity, dtype=int)

        volume = cell.volume
        n_points = mult_i.sum()

        for diffrn in l_diffrn:
            diffrn_orient_matrix = diffrn.diffrn_orient_matrix
            e_up = diffrn_orient_matrix.calc_e_up()
            setup = diffrn.setup
            field = float(setup.field)
            h_loc = (field*e_up[0], field*e_up[1], field*e_up[2])
            diffrn_refln = diffrn.diffrn_refln
            index_h = numpy.array(diffrn_refln.index_h, dtype=int)
            index_k = numpy.array(diffrn_refln.index_k, dtype=int)
            index_l = numpy.array(diffrn_refln.index_l, dtype=int)
            total_peaks += index_h.size
            hkl = (index_h, index_k, index_l)

            f_nucl = crystal.calc_f_nucl(*hkl)
            k_hkl = cell.calc_k_loc(*hkl)
            phase_3d = density_point.calc_phase_3d(hkl, space_group_symop)

            moment_2d, chi_2d_ferro, chi_2d_antiferro = \
                density_point.calc_moment_2d(
                    space_group_symop, cell, atom_site_susceptibility, h_loc,
                    chi_iso_ferro=1., chi_iso_antiferro=1.,
                    flag_two_channel=flag_two_channel)

            chi_ferro = calc_fm_by_density(mult_i, den_ferro_i, n_points,
                                           volume, chi_2d_ferro, phase_3d)
            chi_perp_ferro = calc_moment_perp(k_hkl, chi_ferro)

            chi_aferro = calc_fm_by_density(
                mult_i, den_antiferro_i, n_points, volume, chi_2d_antiferro,
                phase_3d)
            chi_perp_aferro = calc_moment_perp(k_hkl, chi_aferro)

            f_m = calc_fm_by_density(mult_i, den_i, n_points, volume,
                                     moment_2d, phase_3d)
            f_m_perp = calc_moment_perp(k_hkl, f_m)

            f_m_perp_sum = (
                f_m_perp[0] + chi_iso_ferro*chi_perp_ferro[0] +
                chi_iso_antiferro*chi_perp_aferro[0],
                f_m_perp[1] + chi_iso_ferro*chi_perp_ferro[1] +
                chi_iso_antiferro*chi_perp_aferro[1],
                f_m_perp[2] + chi_iso_ferro*chi_perp_ferro[2] +
                chi_iso_antiferro*chi_perp_aferro[2])

            fr_m, delta_fr_m = diffrn.calc_fr(cell, f_nucl, f_m_perp_sum,
                                              delta_f_m_perp=f_m_perp)
            diffrn.diffrn_refln.numpy_fr_calc = fr_m
        for diffrn in l_diffrn:
            diffrn.diffrn_refln.numpy_to_items()
            chi_sq, points = diffrn.diffrn_refln.calc_chi_sq_points()
            refine_ls = RefineLs(goodness_of_fit_all=chi_sq/points,
                                 number_reflns=points)
            diffrn.add_items([refine_ls])

    def maximize_entropy(
            self, c_lambda: float = 1e-3, n_iterations: int = 3000,
            disp: bool = True, d_info: dict = None):
        """Run entropy maximization.

        Arguments
        ---------
            - prior_density: "uniform" (default) or "core"
            - disp: True (default) or False
        """
        crystal = self.crystals()[0]  # FIXME:
        l_diffrn = self.experiments()  # FIXME:
        mem_parameters = self.mem_parameters

        density_point = maximize_entropy(
            crystal, l_diffrn, mem_parameters, c_lambda=c_lambda,
            n_iterations=n_iterations, disp=disp, d_info=d_info)

        self.add_items([density_point])

    def refine_susceptibility(self, disp: bool = True, d_info: dict = None) \
            -> (float, float):
        """Refine susceptibility."""
        crystal = self.crystals()[0]  # FIXME:
        l_diffrn = self.experiments()  # FIXME:
        density_point = self.density_point
        mem_parameters = self.mem_parameters

        # FIXME: temporary solution
        density_point.volume_unit_cell = crystal.cell.volume
        density_point.number_unit_cell = \
            mem_parameters.points_a * mem_parameters.points_b * \
            mem_parameters.points_c

        refine_susceptibility(crystal, l_diffrn, density_point, mem_parameters,
                              disp=disp, d_info=d_info)

    def make_cycle(self, c_lambda: float = 1e-3, n_iterations: int = 3000,
                   n_cycle: int = 10, disp: bool = True, d_info: dict = None):
        """Run Rho - Chi cycle refinement."""
        crystal = self.crystals()[0]  # FIXME:
        l_diffrn = self.experiments()  # FIXME:
        mem_parameters = self.mem_parameters

        density_point = \
            make_cycle(crystal, l_diffrn, mem_parameters, c_lambda=c_lambda,
                       n_iterations=n_iterations, n_cycle=n_cycle, disp=disp,
                       d_info=d_info)
        self.add_items([density_point])

    def calc_moments_in_unit_cell(self, field_loc,
                                  f_name: str = "moments_in_unit_cell.dat"):
        """Calc moments in an unit cell.

        If f_name is given, the result is save in file.
        If it is None, the results are given in output.
        """
        density_point = self.density_point
        crystal = self.crystals()[0]
        mem_parameters = self.mem_parameters
        fract_x, fract_y, fract_z, moment_x, moment_y, moment_z = \
            calc_moments_in_unit_cell(density_point, crystal, mem_parameters,
                                      field_loc)
        if f_name is None:
            return fract_x, fract_y, fract_z, moment_x, moment_y, moment_z
        else:
            ls_out = ["# Moments in an unit cell"]
            ls_out.append("loop_\n_fract_x\n_fract_y\n_fract_z")
            ls_out.append("_moment_x\n_moment_y\n_moment_z")
            for f_x, f_y, f_z, m_x, m_y, m_z in zip(
                    fract_x, fract_y, fract_z, moment_x, moment_y, moment_z):
                ls_out.append(f"{f_x:9.5f} {f_y:9.5f} {f_z:9.5f} \
{m_x:10.3f} {m_y:10.3f} {m_z:10.3f}")
            with open(f_name, "w") as fid:
                fid.write("\n".join(ls_out))

    def calc_sections(self):
        """Calc moments in an unit cell.

        If f_name is given, the result is save in file.
        If it is None, the results are given in output.
        """
        density_point = self.density_point
        crystal = self.crystals()[0]
        loop_section = self.section
        mem_parameters = self.mem_parameters

        for section in loop_section.items:
            calc_section_for_mem(section, density_point, crystal,
                                 mem_parameters)



from typing import NoReturn
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Texture(ItemN):
    """
    Mandatory attributes:
        - g_1
        - g_2
        - h_ax
        - k_ax
        - l_ax

    """
    ATTR_MANDATORY_NAMES = ("g_1", "g_2", "h_ax", "k_ax", "l_ax")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float)
    ATTR_MANDATORY_CIF = ("g_1", "g_2", "h_ax", "k_ax", "l_ax")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("g_1", "g_2", "h_ax", "k_ax", "l_ax")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'g_1': "{:.5f}", 'g_2': "{:.5f}", 'h_ax': "{:.3f}",
                 'k_ax': "{:.3f}", 'l_ax': "{:.3f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "texture"

    def __init__(self, **kwargs) -> NoReturn:
        super(Texture, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class TextureL(LoopN):
    """
    Description of Texture in loop.

    """
    ITEM_CLASS = Texture
    ATTR_INDEX = None
    def __init__(self, loop_name = None) -> NoReturn:
        super(TextureL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name

# s_cont = """
#   loop_
#  _texture_g_1 
#  _texture_g_2 
#  _texture_h_ax 
#  _texture_k_ax 
#  _texture_l_ax 
#  0.1239 0.94211 -0.66119 -0.0541 3.0613
#  0.1239(27) 0.94211(18) -0.66119(41) -0.0541(6) 3.0613(7)
#  """

# obj = TextureL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj[0], end="\n\n")



import numpy
from typing import NoReturn
from cryspy.A_functions_base.function_1_matrices import\
    calc_product_matrices, calc_product_matrix_vector
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.C_item_loop_classes.cl_1_cell import Cell


class PdBackground(ItemN):
    """
    Data items in the REFLN_SUSCEPTIBILITY category record details about the
    susceptibility structure factor tensor and magnetization structure factor 
    tensor calculated for different reflections.

    Mandatory attributes:
        - ub_11, ub_12, ub_13, ub_21, ub_22, ub_23, ub_31, ub_32, ub_33

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """
    ATTR_MANDATORY_NAMES = ("ttheta", "intensity")
    ATTR_MANDATORY_TYPES = (float, float)
    ATTR_MANDATORY_CIF = ("2theta", "intensity")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("intensity", )
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "pd_background"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdBackground, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"ttheta": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class PdBackgroundL(LoopN):
    """
    Description of AtomSite in loop.

    """
    ITEM_CLASS = PdBackground
    ATTR_INDEX = "ttheta"
    def __init__(self, loop_name = None) -> NoReturn:
        super(PdBackgroundL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name
   
    def interpolate_by_points(self, tth):
        l_ttheta = self.ttheta
        l_intensity = self.intensity
        tth_b = numpy.array(l_ttheta, dtype=float)
        int_b = numpy.array(l_intensity, dtype=float)
        if len(l_ttheta) == 0:
            int_1d = numpy.zeros(tth.size, dtype=float)
        else:
            int_1d = numpy.interp(tth, tth_b, int_b)
        return int_1d

# s_cont = """
#  loop_
#  _pd_background_ttheta
#  _pd_background_intensity
#   4.5  256.0
#   40.0  158(2)
#   80.0  65.0 
# """

# obj = PdBackgroundL.from_cif(s_cont)
# print(obj, end="\n\n")



"""Functions for magnetic section.

Functions
---------
    - get_j0_j2_by_symbol
"""
import os
import pycifstar

F_FORMMAG = os.path.join(os.path.dirname(__file__), "formmag.tab")
FORMMAG_data = pycifstar.to_data(F_FORMMAG)


def get_j0_j2_by_symbol(symbol: str):
    """Get <j0>, <j2> parameters by symbol."""
    s_1 = "_atom_type_scat_symbol"
    j0_a2, j0_A1, j0_B1, j0_b2 = 0., 0., 0., 0.
    j0_C1, j0_c2, j0_D = 0., 0., 0.
    j2_a2, j2_A1, j2_B1, j2_b2 = 0., 0., 0., 0.
    j2_C1, j2_c2, j2_D = 0., 0., 0.
    flag_0, flag_2 = False, False
    for loop in FORMMAG_data.loops:
        if "_atom_type_scat_neutron_magnetic_j0_a1" in loop.names:
            hh = [_i1 for _i1, _1 in enumerate(loop[s_1]) if (_1 == symbol)]
            if len(hh) > 0:
                ind = hh[0]
                j0_A1 = float(loop["_atom_type_scat_neutron_magnetic_j0_A1"][
                    ind])
                j0_a2 = float(loop["_atom_type_scat_neutron_magnetic_j0_a2"][
                    ind])
                j0_B1 = float(loop["_atom_type_scat_neutron_magnetic_j0_B1"][
                    ind])
                j0_b2 = float(loop["_atom_type_scat_neutron_magnetic_j0_b2"][
                    ind])
                j0_C1 = float(loop["_atom_type_scat_neutron_magnetic_j0_C1"][
                    ind])
                j0_c2 = float(loop["_atom_type_scat_neutron_magnetic_j0_c2"][
                    ind])
                j0_D = float(loop["_atom_type_scat_neutron_magnetic_j0_D"][
                    ind])
                flag_0 = True
        if "_atom_type_scat_neutron_magnetic_j2_a1" in loop.names:
            hh = [_i1 for _i1, _1 in enumerate(loop[s_1]) if (_1 == symbol)]
            if len(hh) > 0:
                ind = hh[0]
                j2_A1 = float(loop["_atom_type_scat_neutron_magnetic_j2_A1"][
                    ind])
                j2_a2 = float(loop["_atom_type_scat_neutron_magnetic_j2_a2"][
                    ind])
                j2_B1 = float(loop["_atom_type_scat_neutron_magnetic_j2_B1"][
                    ind])
                j2_b2 = float(loop["_atom_type_scat_neutron_magnetic_j2_b2"][
                    ind])
                j2_C1 = float(loop["_atom_type_scat_neutron_magnetic_j2_C1"][
                    ind])
                j2_c2 = float(loop["_atom_type_scat_neutron_magnetic_j2_c2"][
                    ind])
                j2_D = float(loop["_atom_type_scat_neutron_magnetic_j2_D"][
                    ind])
                flag_2 = True
        if all([flag_0, flag_2]):
            break
    return j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D, j2_A1, j2_a2, \
        j2_B1, j2_b2, j2_C1, j2_c2, j2_D



"""Describe AtomSiteSusceptibility, AtomSiteSusceptibilityL."""
from typing import NoReturn
from cryspy.A_functions_base.function_1_atomic_vibrations import \
    vibration_constraints
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class AtomSiteSusceptibility(ItemN):
    """
    AtomSiteSusceptibility class.

    Data items in the ATOM_SITE_MAGNETISM_ANISO category record details about
    magnetic properties of the atoms that occupy the atom sites.

    Mandatory attributes:
        - label
        - type_symbol
        - fract_x
        - fract_y
        - fract_z

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """

    ATTR_MANDATORY_NAMES = ("label", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = (
        "chi_type", "moment_type", "chi_11", "chi_22", "chi_33", "chi_12",
        "chi_13", "chi_23", "moment_11", "moment_22", "moment_33", "moment_12",
        "moment_13", "moment_23")
    ATTR_OPTIONAL_TYPES = (str, str, float, float, float, float, float, float,
                           float, float, float, float, float, float)
    ATTR_OPTIONAL_CIF = (
        "chi_type", "moment_type", "chi_11", "chi_22", "chi_33", "chi_12",
        "chi_13", "chi_23", "moment_11", "moment_22", "moment_33", "moment_12",
        "moment_13", "moment_23")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("chi_11", "chi_22", "chi_33", "chi_12", "chi_13", "chi_23",
                "moment_11", "moment_22", "moment_33", "moment_12",
                "moment_13", "moment_23")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"chi_11": "{:.5f}", "chi_22": "{:.5f}", "chi_33": "{:.5f}",
                 "chi_12": "{:.5f}", "chi_13": "{:.5f}", "chi_23": "{:.5f}",
                 "moment_11": "{:.5f}", "moment_22": "{:.5f}",
                 "moment_33": "{:.5f}", "moment_12": "{:.5f}",
                 "moment_13": "{:.5f}", "moment_23": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"chi_type": ["Ciso", "Cani"],
                     "moment_type": ["Miso", "Mani"]}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "atom_site_susceptibility"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSiteSusceptibility, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def apply_space_group_constraint(self, atom_site, space_group):
        """
        Space group constraints.

        According to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """
        l_numb = atom_site.calc_constr_number(space_group)
        label_aniso = self.label
        label = atom_site.label
        index = label.index(label_aniso)

        flag_chi = self.is_attribute("chi_type")
        if flag_chi:
            flag_chi = self.chi_type.lower().startswith("cani")
        flag_moment = self.is_attribute("moment_type")
        if flag_moment:
            flag_moment = self.moment_type.lower().startswith("mani")
        if flag_chi:
            self.__dict__["chi_11_constraint"] = False
            self.__dict__["chi_22_constraint"] = False
            self.__dict__["chi_33_constraint"] = False
            self.__dict__["chi_12_constraint"] = False
            self.__dict__["chi_13_constraint"] = False
            self.__dict__["chi_23_constraint"] = False
        if flag_moment:
            self.__dict__["moment_11_constraint"] = False
            self.__dict__["moment_22_constraint"] = False
            self.__dict__["moment_33_constraint"] = False
            self.__dict__["moment_12_constraint"] = False
            self.__dict__["moment_13_constraint"] = False
            self.__dict__["moment_23_constraint"] = False
        numb = l_numb[index]

        if flag_chi:
            chi_i = (self.chi_11, self.chi_22, self.chi_33, self.chi_12,
                     self.chi_13, self.chi_23)
            chi_sigma_i = (self.chi_11_sigma, self.chi_22_sigma,
                           self.chi_33_sigma, self.chi_12_sigma,
                           self.chi_13_sigma, self.chi_23_sigma)
            chi_ref_i = (self.chi_11_refinement, self.chi_22_refinement,
                         self.chi_33_refinement, self.chi_12_refinement,
                         self.chi_13_refinement, self.chi_23_refinement)

            chi_i, chi_sigma_i, chi_ref_i, chi_constr_i = \
                vibration_constraints(numb, chi_i, chi_sigma_i, chi_ref_i)

            self.__dict__["chi_11"], self.__dict__["chi_22"], \
                self.__dict__["chi_33"], self.__dict__["chi_12"], \
                self.__dict__["chi_13"], self.__dict__["chi_23"] = chi_i

            self.__dict__["chi_11_sigma"], self.__dict__["chi_22_sigma"], \
                self.__dict__["chi_33_sigma"], self.__dict__["chi_12_sigma"], \
                self.__dict__["chi_13_sigma"], self.__dict__["chi_23_sigma"] =\
                chi_sigma_i

            self.__dict__["chi_11_refinement"], \
                self.__dict__["chi_22_refinement"], \
                self.__dict__["chi_33_refinement"], \
                self.__dict__["chi_12_refinement"], \
                self.__dict__["chi_13_refinement"], \
                self.__dict__["chi_23_refinement"] = chi_ref_i

            self.__dict__["chi_11_constraint"], \
                self.__dict__["chi_22_constraint"], \
                self.__dict__["chi_33_constraint"], \
                self.__dict__["chi_12_constraint"], \
                self.__dict__["chi_13_constraint"], \
                self.__dict__["chi_23_constraint"] = chi_constr_i
        if flag_moment:
            moment_i = (self.moment_11, self.moment_22, self.moment_33,
                        self.moment_12, self.moment_13, self.moment_23)
            moment_sigma_i = (self.moment_11_sigma, self.moment_22_sigma,
                              self.moment_33_sigma, self.moment_12_sigma,
                              self.moment_13_sigma, self.moment_23_sigma)
            moment_ref_i = (
                self.moment_11_refinement, self.moment_22_refinement,
                self.moment_33_refinement, self.moment_12_refinement,
                self.moment_13_refinement, self.moment_23_refinement)

            moment_i, moment_sigma_i, moment_ref_i, moment_constr_i = \
                vibration_constraints(numb, moment_i, moment_sigma_i,
                                      moment_ref_i)

            self.__dict__["moment_11"], self.__dict__["moment_22"], \
                self.__dict__["moment_33"], self.__dict__["moment_12"], \
                self.__dict__["moment_13"], self.__dict__["moment_23"] = \
                moment_i

            self.__dict__["moment_11_sigma"], \
                self.__dict__["moment_22_sigma"], \
                self.__dict__["moment_33_sigma"], \
                self.__dict__["moment_12_sigma"], \
                self.__dict__["moment_13_sigma"], \
                self.__dict__["moment_23_sigma"] = moment_sigma_i

            self.__dict__["moment_11_refinement"], \
                self.__dict__["moment_22_refinement"], \
                self.__dict__["moment_33_refinement"], \
                self.__dict__["moment_12_refinement"], \
                self.__dict__["moment_13_refinement"], \
                self.__dict__["moment_23_refinement"] = moment_ref_i

            self.__dict__["moment_11_constraint"], \
                self.__dict__["moment_22_constraint"], \
                self.__dict__["moment_33_constraint"], \
                self.__dict__["moment_12_constraint"], \
                self.__dict__["moment_13_constraint"], \
                self.__dict__["moment_23_constraint"] = moment_constr_i

    def apply_chi_iso_constraint(self, cell):
        """Isotropic constraint on susceptibility."""
        c_a = cell.cos_a
        s_ib = cell.sin_ib
        s_ig = cell.sin_ig
        c_ib = cell.cos_ib
        c_ig = cell.cos_ig
        # not sure, it is better to check
        if not(self.is_attribute("chi_type")):
            return
        chi_type = self.chi_type
        if chi_type.lower().startswith("ciso"):
            self.__dict__["chi_22"] = self.chi_11
            self.__dict__["chi_33"] = self.chi_11
            self.__dict__["chi_12"] = self.chi_11*c_ig
            self.__dict__["chi_13"] = self.chi_11*c_ib
            self.__dict__["chi_23"] = self.chi_11*(c_ib*c_ig-s_ib*s_ig*c_a)
            self.__dict__["chi_22_sigma"] = self.chi_11_sigma
            self.__dict__["chi_33_sigma"] = self.chi_11_sigma
            self.__dict__["chi_12_sigma"] = self.chi_11_sigma * c_ig
            self.__dict__["chi_13_sigma"] = self.chi_11_sigma * c_ib
            self.__dict__["chi_23_sigma"] = self.chi_11_sigma * \
                (c_ib*c_ig-s_ib*s_ig*c_a)
            self.__dict__["chi_22_refinement"] = False
            self.__dict__["chi_33_refinement"] = False
            self.__dict__["chi_12_refinement"] = False
            self.__dict__["chi_13_refinement"] = False
            self.__dict__["chi_23_refinement"] = False
            self.__dict__["chi_22_constraint"] = True
            self.__dict__["chi_33_constraint"] = True
            self.__dict__["chi_12_constraint"] = True
            self.__dict__["chi_13_constraint"] = True
            self.__dict__["chi_23_constraint"] = True

    def apply_moment_iso_constraint(self, cell):
        """Isotropic constraint on moment."""
        c_a = cell.cos_a
        s_ib = cell.sin_ib
        s_ig = cell.sin_ig
        c_ib = cell.cos_ib
        c_ig = cell.cos_ig
        # not sure, it is better to check
        if not(self.is_attribute("moment_type")):
            return
        moment_type = self.moment_type
        if moment_type.lower().startswith("miso"):
            self.__dict__["moment_22"] = self.moment_11
            self.__dict__["moment_33"] = self.moment_11
            self.__dict__["moment_12"] = self.moment_11 * c_ig
            self.__dict__["moment_13"] = self.moment_11 * c_ib
            self.__dict__["moment_23"] = self.moment_11 * \
                (c_ib*c_ig-s_ib*s_ig*c_a)

            self.__dict__["moment_22_sigma"] = self.moment_11_sigma
            self.__dict__["moment_33_sigma"] = self.moment_11_sigma
            self.__dict__["moment_12_sigma"] = self.moment_11_sigma * c_ig
            self.__dict__["moment_13_sigma"] = self.moment_11_sigma * c_ib
            self.__dict__["moment_23_sigma"] = self.moment_11_sigma * \
                (c_ib*c_ig-s_ib*s_ig*c_a)

            self.__dict__["moment_22_refinement"] = False
            self.__dict__["moment_33_refinement"] = False
            self.__dict__["moment_12_refinement"] = False
            self.__dict__["moment_13_refinement"] = False
            self.__dict__["moment_23_refinement"] = False
            self.__dict__["moment_22_constraint"] = True
            self.__dict__["moment_33_constraint"] = True
            self.__dict__["moment_12_constraint"] = True
            self.__dict__["moment_13_constraint"] = True
            self.__dict__["moment_23_constraint"] = True


class AtomSiteSusceptibilityL(LoopN):
    """
    Description of AtomSite in loop.

    Methods
    -------
        - apply_space_group_constraint
        - apply_chi_iso_constraint
        - apply_moment_iso_constraint
    """

    ITEM_CLASS = AtomSiteSusceptibility
    ATTR_INDEX = "label"

    def __init__(self, loop_name=None) -> NoReturn:
        super(AtomSiteSusceptibilityL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name

    def apply_space_group_constraint(self, atom_site, space_group):
        """Apply space group constraint."""
        for item in self.items:
            item.apply_space_group_constraint(atom_site, space_group)

    def apply_chi_iso_constraint(self, cell):
        """Apply isotropic constraint on susceptibility."""
        for item in self.items:
            item.apply_chi_iso_constraint(cell)

    def apply_moment_iso_constraint(self, cell):
        """Apply isotropic constraint on moments."""
        for item in self.items:
            item.apply_moment_iso_constraint(cell)

# s_cont = """
#  loop_
#  _atom_site_susceptibility_label
#  _atom_site_susceptibility_chi_type
#  _atom_site_susceptibility_chi_11
#  _atom_site_susceptibility_chi_12
#  _atom_site_susceptibility_chi_13
#  _atom_site_susceptibility_chi_22
#  _atom_site_susceptibility_chi_23
#  _atom_site_susceptibility_chi_33
#  _atom_site_susceptibility_moment_type
#  _atom_site_susceptibility_moment_11
#  _atom_site_susceptibility_moment_12
#  _atom_site_susceptibility_moment_13
#  _atom_site_susceptibility_moment_22
#  _atom_site_susceptibility_moment_23
#  _atom_site_susceptibility_moment_33
#   Fe3A Cani -3.468(74) 0.0 0.0 -3.468 0.0 -3.468 Mani 0. 0. 0. 0. 0. 0.
#   Fe3B Cani 3.041      0.0 0.0  3.041 0.0  3.041 Mani 0. 0. 0. 0. 0. 0.
#   """

# obj = AtomSiteSusceptibilityL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["Fe3A"], end="\n\n")



"""Transform data to objects of cryspy library."""
import os
import os.path
from typing import List

from pycifstar import Global, to_global, Items, Loop, Data

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSite, \
    AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_type import AtomType, \
    AtomTypeL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import \
    AtomSiteAniso, AtomSiteAnisoL
from cryspy.C_item_loop_classes.cl_1_atom_site_moment import \
    AtomSiteMoment, AtomSiteMomentL
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import \
    AtomSiteSusceptibility, AtomSiteSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_atom_type_scat import \
    AtomTypeScat, AtomTypeScatL
from cryspy.C_item_loop_classes.cl_1_atom_local_axes import \
    AtomLocalAxes, AtomLocalAxesL
from cryspy.C_item_loop_classes.cl_1_atom_electron_configuration \
    import AtomElectronConfiguration, AtomElectronConfigurationL
from cryspy.C_item_loop_classes.cl_1_cell import Cell, CellL
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2, Chi2L
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation, DiffrnRadiationL
from cryspy.C_item_loop_classes.cl_1_diffrn_refln import DiffrnRefln, \
    DiffrnReflnL
from cryspy.C_item_loop_classes.cl_1_exclude import Exclude, ExcludeL
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction, \
    ExtinctionL
from cryspy.C_item_loop_classes.cl_1_phase import Phase, PhaseL
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_reflex_asymmetry import\
    Pd2dInstrReflexAsymmetry, Pd2dInstrReflexAsymmetryL
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_resolution import \
    Pd2dInstrResolution, Pd2dInstrResolutionL
from cryspy.C_item_loop_classes.cl_1_pd2d_meas import Pd2dMeas
from cryspy.C_item_loop_classes.cl_1_pd2d_proc import Pd2dProc
from cryspy.C_item_loop_classes.cl_1_pd2d_peak import Pd2dPeak, \
    Pd2dPeakL
from cryspy.C_item_loop_classes.cl_1_pd2d_background import \
    Pd2dBackground
from cryspy.C_item_loop_classes.cl_1_pd_background import \
    PdBackground, PdBackgroundL
from cryspy.C_item_loop_classes.cl_1_pd_instr_reflex_asymmetry import \
    PdInstrReflexAsymmetry, PdInstrReflexAsymmetryL
from cryspy.C_item_loop_classes.cl_1_pd_instr_resolution import \
    PdInstrResolution, PdInstrResolutionL
from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeas, PdMeasL
from cryspy.C_item_loop_classes.cl_1_pd_proc import PdProc, PdProcL
from cryspy.C_item_loop_classes.cl_1_pd_peak import PdPeak, PdPeakL
from cryspy.C_item_loop_classes.cl_1_range import Range, RangeL
from cryspy.C_item_loop_classes.cl_1_refln import Refln, ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibility, ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs, \
    RefineLsL
from cryspy.C_item_loop_classes.cl_1_setup import Setup, SetupL
from cryspy.C_item_loop_classes.cl_1_space_group_symop_magn_centering import \
    SpaceGroupSymopMagnCentering, SpaceGroupSymopMagnCenteringL
from cryspy.C_item_loop_classes.cl_1_texture import Texture, TextureL
from cryspy.C_item_loop_classes.cl_1_mem_parameters import \
    MEMParameters, MEMParametersL

from cryspy.C_item_loop_classes.cl_2_atom_site_scat import \
    AtomSiteScat, AtomSiteScatL
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import \
    DiffrnOrientMatrix, DiffrnOrientMatrixL
from cryspy.C_item_loop_classes.cl_2_section import Section, SectionL
from cryspy.C_item_loop_classes.cl_2_space_group import SpaceGroup
from cryspy.C_item_loop_classes.cl_2_space_group_symop_magn_operation import \
    SpaceGroupSymopMagnOperation, SpaceGroupSymopMagnOperationL

from cryspy.C_item_loop_classes.cl_3_density_point import DensityPoint, \
    DensityPointL

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn
from cryspy.E_data_classes.cl_2_pd import Pd
from cryspy.E_data_classes.cl_2_pd2d import Pd2d

from cryspy.G_global_classes.cl_1_rhochi import RhoChi
from cryspy.G_global_classes.cl_1_mem import MEM


L_ITEM_CLASS = []
L_LOOP_CLASS = []
L_DATA_CLASS = []
L_GLOBAL_CLASS = []
L_FUNCTION_ALL = []

L_ITEM_CLASS.extend([
    AtomElectronConfiguration, AtomLocalAxes,
    AtomSite, AtomSiteAniso, AtomSiteMoment, AtomSiteScat,
    AtomSiteSusceptibility, AtomType, AtomTypeScat,
    Cell, Chi2,
    DensityPoint, DiffrnRadiation, DiffrnOrientMatrix, DiffrnRefln,
    Exclude, Extinction, MEMParameters,
    Phase, PdBackground,
    PdInstrReflexAsymmetry, PdInstrResolution, PdMeas, PdProc, PdPeak,
    Pd2dBackground, Pd2dInstrReflexAsymmetry,
    Pd2dInstrResolution, Pd2dMeas, Pd2dProc, Pd2dPeak,
    Range, RefineLs, Refln, ReflnSusceptibility,
    Setup, SpaceGroupSymopMagnCentering, SpaceGroupSymopMagnOperation,
    SpaceGroup, Section, Texture])

L_LOOP_CLASS.extend([
    AtomLocalAxesL, AtomElectronConfigurationL,
    AtomSiteL, AtomSiteAnisoL, AtomSiteMomentL, AtomSiteScatL,
    AtomSiteSusceptibilityL, AtomTypeL, AtomTypeScatL,
    CellL, Chi2L,
    DensityPointL, DiffrnRadiationL, DiffrnOrientMatrixL, DiffrnReflnL,
    ExtinctionL, ExcludeL, MEMParametersL,
    PhaseL, PdBackgroundL, PdInstrReflexAsymmetryL, PdInstrResolutionL,
    PdMeasL, PdProcL, PdPeakL,
    Pd2dInstrReflexAsymmetryL, Pd2dInstrResolutionL, Pd2dPeakL,
    RangeL, RefineLsL, ReflnL, ReflnSusceptibilityL,
    SectionL, SetupL, SpaceGroupSymopMagnCenteringL,
    SpaceGroupSymopMagnOperationL, TextureL])

L_DATA_CLASS.extend([Crystal, MagCrystal, Diffrn, Pd, Pd2d])

L_GLOBAL_CLASS.extend([MEM, RhoChi])


def is_in_rcif_block(rcif_block, cryspy_obj):
    """Is in rcif block."""
    ls_cryspy = cryspy_obj.to_cif.split("\n")
    l_name = [_.strip().split()[0] for _ in ls_cryspy if _.startswith("_")]
    l_flag = [rcif_block.is_value(_name) for _name in l_name]
    return any(l_flag), all(l_flag)


def file_to_globaln(file_name: str, item_classes=(), loop_classes=(),
                    data_classes=(), global_classes=()) -> GlobalN:
    """Transfer file content of cif format to cryspy objects."""
    if not(os.path.isfile(file_name)):
        raise UserWarning(f"File '{file_name:}' is not found")
        return
    s_cont = str(to_global(file_name))
    res = str_to_globaln(s_cont, item_classes=item_classes,
                         loop_classes=loop_classes, data_classes=data_classes,
                         global_classes=global_classes)
    return res


def str_to_items(s_cont: str, item_classes=(), loop_classes=()) -> list:
    """Transfer string of cif format to cryspy objects."""
    data_cif = Data()
    data_cif.take_from_string(s_cont)
    l_item_class = L_ITEM_CLASS + list(item_classes)
    l_loop_class = L_LOOP_CLASS + list(loop_classes)

    l_item = items_to_itemsn(data_cif.items, l_item_class)
    for loop_cif in data_cif.loops:
        l_loop_obj = loop_to_loopn(loop_cif, l_loop_class)
        l_item.extend(l_loop_obj)
    return l_item


def str_to_globaln(s_cont: str, item_classes=(), loop_classes=(),
                   data_classes=(), global_classes=()) -> GlobalN:
    """Transfer string of cif format to cryspy objects."""
    global_cif = Global()
    global_cif.take_from_string(s_cont)
    l_global_item = []
    l_cls_global = []
    l_item_class = L_ITEM_CLASS + list(item_classes)
    l_loop_class = L_LOOP_CLASS + list(loop_classes)
    l_data_class = L_DATA_CLASS + list(data_classes)
    l_global_class = L_GLOBAL_CLASS + list(global_classes)
    if len(global_cif.items)+len(global_cif.loops) > 0:
        l_item_obj = items_to_itemsn(global_cif.items, l_item_class)
        l_global_item.extend(l_item_obj)

        for item in l_item_obj:
            if type(item) not in l_cls_global:
                l_cls_global.append(type(item))

        for loop_cif in global_cif.loops:
            l_loop_obj = loop_to_loopn(loop_cif, l_loop_class)
            l_global_item.extend(l_loop_obj)
            for loop_obj in l_loop_obj:
                if type(loop_obj) not in l_cls_global:
                    l_cls_global.append(type(loop_obj))

    for data_cif in global_cif.datas:
        flag = False
        str_data = str(data_cif)
        for cls_data in l_data_class:
            # FIXME: it's bad solution as we loose information which are
            #        not specified as knonw in cryspy library.
            data_obj = cls_data.from_cif(str_data)
            if data_obj is not None:
                l_global_item.append(data_obj)
                if not(cls_data in l_cls_global):
                    l_cls_global.append(cls_data)
                flag = True
        if not(flag):
            l_data_item = []
            l_cls_data = []

            l_item_obj = items_to_itemsn(data_cif.items, l_item_class)
            l_data_item.extend(l_item_obj)
            for item in l_item_obj:
                if type(item) not in l_cls_data:
                    l_cls_data.append(type(item))

            for loop_cif in data_cif.loops:
                l_loop_obj = loop_to_loopn(loop_cif, l_loop_class)
                l_data_item.extend(l_loop_obj)
                for loop_obj in l_loop_obj:
                    if type(loop_obj) not in l_cls_data:
                        l_cls_data.append(type(loop_obj))

            if len(l_data_item) > 0:
                data_obj = items_to_datan(data_cif.name, l_data_item,
                                          l_data_class)
                l_global_item.append(data_obj)

    global_obj = items_to_globaln(global_cif.name, l_global_item,
                                  l_global_class)
    return global_obj


def find_name_prefix(loop_name, l_loop_prefix):
    """Find name prefix."""
    loop_name = loop_name.strip("_")
    l_loop = [loop_prefix for loop_prefix in l_loop_prefix if
              loop_name.startswith(loop_prefix)]
    if len(l_loop) > 0:
        l_loop.sort(key=lambda x: len(x))
        prefix = l_loop[-1]
        name = loop_name.strip("_")[len(prefix)+1:]
        separator = loop_name.strip("_")[len(prefix)]
        return prefix, name, separator
    return None, None, None


# TODO: l_item_class should be added as not necessary l_loop_class fully
#       correspond to l_item_class
def loop_to_loopn(loop_cif: Loop, l_loop_class: list) -> LoopN:
    """Transform Loop object to ItemN objects."""
    # l_item_class for future.
    l_item_class = []
    l_item_prefix = [item_cls.PREFIX for item_cls in l_item_class]
    l_loop_prefix = [loop_cls.ITEM_CLASS.PREFIX for loop_cls in l_loop_class]
    l_loop_name = [name.lower() for name in loop_cif.names]
    flag_id = False
    flag_label = False
    flag_type = False
    l_loop_prefix_in = []
    l_loop_name_in = []
    l_loop_separator_in = []
    for loop_name in l_loop_name:
        prefix, name, separator = find_name_prefix(loop_name,
                                                   l_loop_prefix)
        if prefix is not None:
            l_name_allowed = [h.lower() for h in l_loop_class[
                l_loop_prefix.index(prefix)].ITEM_CLASS.ATTR_CIF]
            if not(name.lower() in l_name_allowed):
                prefix = None
                name = None
        if (prefix is None) | (name is None):
            prefix, name, separator = find_name_prefix(loop_name,
                                                       l_item_prefix)
            if prefix is not None:
                l_name_allowed = [h.lower() for h in l_item_class[
                    l_item_prefix.index(prefix)].ATTR_CIF]
                if not(name.lower() in l_name_allowed):
                    prefix = None
                    name = None
        if (prefix is None) | (name is None):
            if loop_name.find(".") != -1:
                prefix, name = loop_name.strip("_").split(".")[:2]
                separator = "."
            else:
                prefix = loop_name.strip("_").split("_")[0]
                name = loop_name.strip("_")[len(prefix)+1:]
                separator = "_"
        if name.lower() == "id":
            flag_id = True
            prefix_id = prefix
            loop_name_id = loop_name
        if name.lower() == "label":
            flag_label = True
            prefix_label = prefix
            loop_name_label = loop_name
        if name.lower() == "type":
            flag_type = True
            prefix_type = prefix
            loop_name_type = loop_name
        l_loop_prefix_in.append(prefix)
        l_loop_name_in.append(name)
        l_loop_separator_in.append(separator)


    s_loop_prefix_in = set(l_loop_prefix_in)
    l_loopn = []
    for prefix in s_loop_prefix_in:
        item_cls = ItemN
        loop_cls = LoopN
        if prefix in l_loop_prefix:
            loop_cls = l_loop_class[l_loop_prefix.index(prefix)]
        elif prefix in l_item_prefix:
            # for future
            item_cls = l_item_class[l_loop_prefix.index(prefix)]
        l_loop_name_cif = []
        ll_val_cif = []
        separator = "_"
        for loop_name, prefix_in, name_in, separator_in in zip(
                l_loop_name, l_loop_prefix_in, l_loop_name_in,
                l_loop_separator_in):
            if prefix == prefix_in:
                l_loop_name_cif.append(loop_name)
                ll_val_cif.append(loop_cif[loop_name])
                separator = separator_in
        if flag_id:
            if prefix_id != prefix:
                loop_new_name_id = "_"+prefix+f"{separator:}id"
                l_loop_name_cif.append(loop_new_name_id)
                ll_val_cif.append(loop_cif[loop_name_id])
        elif flag_label:
            if prefix_label != prefix:
                loop_new_name_label = "_"+prefix+f"{separator:}label"
                l_loop_name_cif.append(loop_new_name_label)
                ll_val_cif.append(loop_cif[loop_name_label])
        elif flag_type:
            if prefix_type != prefix:
                loop_new_name_type = "_"+prefix+f"{separator:}type"
                l_loop_name_cif.append(loop_new_name_type)
                ll_val_cif.append(loop_cif[loop_name_type])
        ll_val_cif_t = [[ll_val_cif[i_2][i_1] for i_2 in range(len(
            ll_val_cif))] for i_1 in range(len(ll_val_cif[0]))]
        obj = Loop(names=l_loop_name_cif, values=ll_val_cif_t)
        loopn = loop_cls.from_cif(str(obj))
        if loopn is None:
            loop_cls = LoopN
            loopn = loop_cls.from_cif(str(obj))

        if not(loopn is None):
            l_loopn.append(loopn)
        else:
            print("l_loop_name_cif: \n", l_loop_name_cif)
    return l_loopn


def items_to_itemsn(items_cif: Items, l_item_class: list) -> List[ItemN]:
    """Transform Items object to items of ItemN classes."""
    l_item = []
    l_name_cif = list(items_cif.names)
    flag_point_separator = all([name.find(".") == -1
                                for name in l_name_cif])
    if flag_point_separator:
        separator = "_"
    else:
        separator = "."

    for cls_item in l_item_class:
        prefix = cls_item.PREFIX

        separator_loc = "_"

        l_cls_mand_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                          for attr_cif in cls_item.ATTR_MANDATORY_CIF]

        l_cls_all_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                         for attr_cif in cls_item.ATTR_MANDATORY_CIF]

        flag_mand = all([items_cif.is_value(name) for name in l_cls_mand_cif])
        flag_any = any([items_cif.is_value(name) for name in l_cls_all_cif])

        if not(flag_any):
            separator_loc = "."
            l_cls_mand_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                              for attr_cif in cls_item.ATTR_MANDATORY_CIF]

            l_cls_all_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                             for attr_cif in cls_item.ATTR_MANDATORY_CIF]

            flag_mand = all([items_cif.is_value(name) for name in
                             l_cls_mand_cif])
            flag_any = any([items_cif.is_value(name) for name in
                            l_cls_all_cif])

        item_obj = None
        if flag_mand:
            items_small = items_cif.items_with_prefix(f"_{prefix:}")
            item_obj = cls_item.from_cif(str(items_small))
        elif flag_any:
            items_small = items_cif.items_with_prefix(f"_{prefix:}")
            item_obj = ItemN.from_cif(str(items_small))

        if item_obj is not None:
            l_item.append(item_obj)
            l_name_cif_remove = [name.lower() for name in l_cls_all_cif
                                 if bool(l_name_cif.count(name.lower()))]
            for name in l_name_cif_remove:
                l_name_cif.remove(name)

    if len(l_name_cif) != 0:
        if separator == ".":
            l_prefix_cif = [(name[1:]).split(separator)[0]
                            for name in l_name_cif]
        else:
            l_prefix_cif = find_prefixes(l_name_cif)

        for prefix in l_prefix_cif:
            items_small = items_cif.items_with_prefix(f"_{prefix:}")
            item_obj = ItemN.from_cif(str(items_small))
            l_item.append(item_obj)
    return l_item


def find_prefixes(l_name_cif):
    """Find prefix."""
    l_prefix_cif = list(set([(name[1:]).split("_")[0] for name in l_name_cif]))
    return l_prefix_cif


def items_to_globaln(global_name: str, items: list, l_global_class):
    """
    Items to suitable GlobalN class.

    Parameters
    ----------
    global_name : str
        name of the GlobalN.
    items : List[Union[ItemN, LoopN, DataN]]
        List of items.
    l_global_class : list
        List of GlobalN classes to choose one of them.

    Returns
    -------
    global_obj : TYPE
        Object of GlobalN class or one of l_global_classes.

    """
    global_obj = None
    classes = set([type(item) for item in items])
    for cls_global in l_global_class:
        flag_mand = all([cls_mand in classes
                         for cls_mand in cls_global.CLASSES_MANDATORY])
        flag_opt = all([class_ in cls_global.CLASSES
                        for class_ in classes])
        if (flag_mand & flag_opt):
            global_obj = cls_global()
            global_obj.global_name = global_name
            global_obj.add_items(items)
    if global_obj is None:
        global_obj = GlobalN.make_container((), tuple(classes), "global")
        global_obj.global_name = global_name
        global_obj.add_items(items)
    return global_obj


def items_to_datan(data_name: str, items: list, l_data_class):
    """
    Items to suitable GlobalN class.

    Parameters
    ----------
    data_name : str
        name of the DataN.
    items : List[Union[ItemN, LoopN]]
        List of items.
    l_data_class : list
        List of DataN classes to choose one of them.

    Returns
    -------
    data_obj : TYPE
        Object of DataN class or one of l_data_classes.

    """
    data_obj = None
    classes = set([type(item) for item in items])
    for cls_data in l_data_class:
        flag_mand = all([cls_mand in classes
                         for cls_mand in cls_data.CLASSES_MANDATORY])
        flag_opt = all([class_ in cls_data.CLASSES
                        for class_ in classes])
        if (flag_mand & flag_opt):
            data_obj = cls_data()
            data_obj.data_name = data_name
            data_obj.add_items(items)

    if data_obj is None:
        data_obj = DataN.make_container((), tuple(classes), "data")
        data_obj.data_name = data_name
        data_obj.add_items(items)
    return data_obj

# s_cont = """
# global_
#   _unknown_fuck_d 7
#   _unknown_fuck_e 3546
#   _unknown_fuck_type_g 14
#   _test_fuck_d 7
#   _test_fuck_e 3546
#   _test_fuck_type_g 14

#   data_Fe3O4

#   _test2_fuck_d 7
#   _test2_fuck_e 3546
#   _test2_fuck_type_g 14

#   _cell_angle_alpha 90.0
#   _cell_angle_beta 90.0
#   _cell_angle_gamma 90.0
#   _cell_length_a 8.56212()
#   _cell_length_b 8.56212
#   _cell_length_c 8.56212
#   _space_group_it_coordinate_system_code 2
#   _space_group_IT_number    227
#   loop_
#   _atom_site_adp_type
#   _atom_site_B_iso_or_equiv
#   _atom_site_fract_x
#   _atom_site_fract_y
#   _atom_site_fract_z
#   _atom_site_label
#   _atom_site_occupancy
#   _atom_site_type_symbol
#   Uani 0.0 0.125 0.125 0.125 Fe3A 1.0 Fe3+
#   Uani 0.0 0.5 0.5 0.5 Fe3B 1.0 Fe3+
#   Uiso 0.0 0.25521 0.25521 0.25521 O1 1.0 O2-
#   loop_
#   _atom_type_scat_length_neutron
#   _atom_type_symbol
#     0.945 Fe3+
#   0.5803 O2-
#   loop_
#   _atom_site_aniso_U_11
#   _atom_site_aniso_U_12
#   _atom_site_aniso_U_13
#   _atom_site_aniso_U_22
#   _atom_site_aniso_U_23
#   _atom_site_aniso_U_33
#   _atom_site_aniso_label
#   0.0 0.0 0.0 0.0 0.0 0.0 Fe3A
#   0.0 0.0 0.0 0.0 0.0 0.0 Fe3B
#   loop_
#   _atom_site_scat_label
#   _atom_site_scat_lande
#   Fe3A 2.0
#   Fe3B 2.0
#   loop_
#   _atom_site_susceptibility_label
#   _atom_site_susceptibility_chi_type
#   _atom_site_susceptibility_chi_11
#   _atom_site_susceptibility_chi_12
#   _atom_site_susceptibility_chi_13
#   _atom_site_susceptibility_chi_22
#   _atom_site_susceptibility_chi_23
#   _atom_site_susceptibility_chi_33
#   Fe3A Cani -3.468(74) 0.0 0.0 -3.468 0.0 -3.468
#   Fe3B Cani 3.041      0.0 0.0  3.041 0.0  3.041
#   data_mono
#   _setup_wavelength     0.840
#   _setup_field          1.000
#   _diffrn_radiation_polarization 1.0
#   _diffrn_radiation_efficiency   1.0
#   _extinction_mosaicity 100.0
#   _extinction_radius    50.0
#   _extinction_model     gauss
#   _diffrn_orient_matrix_UB_11 6.59783
#   _diffrn_orient_matrix_UB_12 -6.99807
#   _diffrn_orient_matrix_UB_13 3.3663
#   _diffrn_orient_matrix_UB_21 2.18396
#   _diffrn_orient_matrix_UB_22 -2.60871
#   _diffrn_orient_matrix_UB_23 -9.5302
#   _diffrn_orient_matrix_UB_31 7.4657
#   _diffrn_orient_matrix_UB_32 6.94702
#   _diffrn_orient_matrix_UB_33 -0.18685
#   _phase_label  Fe3O4
#   loop_
#   _diffrn_refln_index_h
#   _diffrn_refln_index_k
#   _diffrn_refln_index_l
#   _diffrn_refln_fr
#   _diffrn_refln_fr_sigma
#   0 0 8 0.64545 0.01329
#   2 0 6 1.75682 0.04540
#   0 2 6 1.67974 0.03711
# """
# obj = str_to_globaln(s_cont)
# for var_name in obj.get_variable_names():
#     print(var_name)
# obj



"""Description of Pd class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
from warnings import warn
import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_cos_ang
from cryspy.A_functions_base.function_1_matrices import calc_mRmCmRT

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN

from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_pd_background import PdBackgroundL
from cryspy.C_item_loop_classes.cl_1_pd_instr_reflex_asymmetry import \
    PdInstrReflexAsymmetry
from cryspy.C_item_loop_classes.cl_1_pd_instr_resolution import \
    PdInstrResolution
from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeasL
from cryspy.C_item_loop_classes.cl_1_pd_proc import PdProcL
from cryspy.C_item_loop_classes.cl_1_pd_peak import PdPeakL
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_texture import Texture
from cryspy.C_item_loop_classes.cl_1_exclude import ExcludeL

from cryspy.E_data_classes.cl_1_crystal import Crystal


class Pd(DataN):
    """
    Pd class.

    Data items in the DIFFRN category record details about
    powder diffraction measurements.

    Methods
    -------
        - calc_profile
        - calc_chi_sq
        - simmulation
        - calc_iint 
        - _gauss_pd
        - _lor_pd
        - calc_shape_profile
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - apply_constraints

    Attributes
    ----------
        - setup, pd_instr_resolution, phase, pd_background, pd_meas (mandatory)
        - diffrn_radiation, chi2, range, extinction, pd_instr_reflex_asymmetry
          texture, exclude, pd_proc, pd_peak, refine_ls, refln_#phase_name
          refln_susceptibility_#phase_name (optional)
    """

    CLASSES_MANDATORY = (Setup, PdInstrResolution, PhaseL, PdBackgroundL,
                         PdMeasL)
    CLASSES_OPTIONAL = (DiffrnRadiation, Chi2, Range, Extinction,
                        PdInstrReflexAsymmetry, Texture, ExcludeL, PdProcL,
                        PdPeakL, RefineLs, ReflnL, ReflnSusceptibilityL)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "pd"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Pd, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_profile(self, tth, l_crystal, l_peak_in=None, l_refln_in=None,
                     l_refln_susceptibility_in=None, l_dd_in=None,
                     flag_internal: bool = True, flag_polarized: bool = True):
        """Calculate intensity for the given diffraction angles.

        Arguments
        ---------
            - tth: 1D numpy array of 2theta in degrees
            - l_crystal: a list of Crystal objects of cryspy library
            - l_peak_in: precalculated data about integrated intensities
              (used to speed up the calculations)
            - l_refln_in: precalculated data about nuclear structure factors
              (used to speed up the calculations)
            - l_refln_susceptibility_in: precalculated data about
              (used to speed up the calculations)
            - flag_internal: a flag to calculate or to use internal objects.
                   It should be True if user call the function.
                   It's True by default.

        Output
        ------
            - proc: output profile
            - l_peak: data about peaks
            - l_refln: data about nuclear structure factor
        """
        proc = PdProcL()
        proc.numpy_ttheta = tth

        background = self.pd_background
        if background is not None:
            int_bkgd = background.interpolate_by_points(tth)
        else:
            int_bkgd = 0.*tth

        proc.numpy_intensity_bkg_calc = int_bkgd

        wavelength = float(self.setup.wavelength)

        tth_min = tth.min()
        tth_max = tth.max() + 3.
        if tth_max > 180.:
            tth_max = 180.
        sthovl_min = numpy.sin(0.5*tth_min*numpy.pi/180.)/wavelength
        sthovl_max = numpy.sin(0.5*tth_max*numpy.pi/180.)/wavelength

        texture = self.texture
        if texture is not None:
            h_ax, k_ax, l_ax = texture.h_ax, texture.k_ax, texture.l_ax
            g_1, g_2 = texture.g_1, texture.g_2

        res_u_1d = numpy.zeros(tth.shape[0], dtype=float)
        res_d_1d = numpy.zeros(tth.shape[0], dtype=float)

        phase = self.phase
        _h = len(phase.label)
        if l_peak_in is None:
            l_peak_in = len(phase.label)*[None]
        else:
            if _h != len(l_peak_in):
                l_peak_in = len(phase.label)*[None]
        if l_refln_in is None:
            l_refln_in = len(phase.label)*[None]
        else:
            if len(phase.label) != len(l_refln_in):
                l_refln_in = len(phase.label)*[None]
        if l_refln_susceptibility_in is None:
            l_refln_susceptibility_in = len(phase.label) * [None]
        else:
            if _h != len(l_refln_susceptibility_in):
                l_refln_susceptibility_in = len(phase.label) * [None]

        if l_dd_in is None:
            l_dd_in = len(phase.label) * [None]
        else:
            if _h != len(l_dd_in):
                l_dd_in = len(phase.label) * [None]

        l_dd_out = []
        l_peak, l_refln, l_refln_s = [], [], []
        for phase_item, peak_in, refln_in, refln_susceptibility_in, dd_in in \
                zip(phase.items, l_peak_in, l_refln_in,
                    l_refln_susceptibility_in, l_dd_in):
            phase_label = phase_item.label
            phase_scale = phase_item.scale
            phase_igsize = phase_item.igsize

            dd_out = {}
            for i_crystal, crystal in enumerate(l_crystal):
                if crystal.data_name.lower() == phase_label.lower():
                    ind_cry = i_crystal
                    break
            if ind_cry is None:
                warn(f"Crystal with name '{phase_label:}' is not found.",
                     UserWarning)
                return
            crystal = l_crystal[ind_cry]
            scale = phase_scale
            i_g = phase_igsize

            if peak_in is not None:
                index_h = peak_in.numpy_index_h
                index_k = peak_in.numpy_index_k
                index_l = peak_in.numpy_index_l
                mult = peak_in.numpy_index_mult
            else:
                if texture is None:
                    index_h, index_k, index_l, mult = crystal.calc_hkl(
                        sthovl_min, sthovl_max)
                else:
                    index_h, index_k, index_l, mult = \
                        crystal.calc_hkl_in_range(sthovl_min, sthovl_max)

            peak = PdPeakL(loop_name=phase_label)

            cond_1 = not(crystal.is_variables())
            cond_2 = (peak_in is not None) & (refln_in is not None)
            if cond_1 & cond_2:
                np_iint_u = peak_in.numpy_intensity_up
                np_iint_d = peak_in.numpy_intensity_down
                refln = refln_in
                refln_s = refln_susceptibility_in
            else:
                np_iint_u, np_iint_d, refln, refln_s = self.calc_iint(
                    index_h, index_k, index_l, crystal,
                    flag_internal=flag_internal)
                refln.loop_name = phase_label
                refln_s.loop_name = phase_label

            l_refln.append(refln)
            l_refln_s.append(refln_s)

            cell = crystal.cell
            sthovl_hkl = cell.calc_sthovl(index_h, index_k, index_l)

            tth_hkl_rad = numpy.where(sthovl_hkl*wavelength < 1.,
                                      2.*numpy.arcsin(sthovl_hkl*wavelength),
                                      numpy.pi)
            tth_hkl = tth_hkl_rad*180./numpy.pi

            profile_2d, tth_zs, h_pv = self.calc_shape_profile(
                tth, tth_hkl, i_g)

            peak.numpy_index_h = index_h
            peak.numpy_index_k = index_k
            peak.numpy_index_l = index_l
            peak.numpy_index_mult = mult
            peak.numpy_intensity_up = np_iint_u
            peak.numpy_intensity_down = np_iint_d
            peak.numpy_ttheta = tth_hkl+self.setup.offset_ttheta
            peak.numpy_width_ttheta = h_pv

            np_iint_u_2d = numpy.meshgrid(tth, np_iint_u*mult,
                                          indexing="ij")[1]
            np_iint_d_2d = numpy.meshgrid(tth, np_iint_d*mult,
                                          indexing="ij")[1]

            # texture
            if texture is not None:
                cond_1 = dd_in is not None
                cond_2 = ((not(texture.is_variables())) &
                          (not(cell.is_variables())))
                if cond_1 & cond_2:
                    texture_2d = dd_in["texture_2d"]
                else:
                    cos_alpha_ax = calc_cos_ang(cell, h_ax, k_ax, l_ax,
                                                index_h, index_k, index_l)
                    c_help = 1.-cos_alpha_ax**2
                    c_help[c_help < 0.] = 0.
                    sin_alpha_ax = numpy.sqrt(c_help)
                    cos_alpha_ax_2d = numpy.meshgrid(tth, cos_alpha_ax,
                                                     indexing="ij")[1]
                    sin_alpha_ax_2d = numpy.meshgrid(tth, sin_alpha_ax,
                                                     indexing="ij")[1]
                    # cos_alpha_2d = cos_alpha_ax_2d*cos_alpha_ang_2d + \
                    #                sin_alpha_ax_2d*sin_alpha_ang_2d
                    # cos_alpha_ang_2d, sin_alpha_ang_2d
                    cos_alpha_2d = cos_alpha_ax_2d*1.+sin_alpha_ax_2d*0.
                    texture_2d = g_2 + (1. - g_2) * (1./g_1 + (g_1**2 - 1./g_1)
                                                     * cos_alpha_2d**2)**(-1.5)

                dd_out["texture_2d"] = texture_2d
                profile_2d = profile_2d*texture_2d

            res_u_2d = profile_2d*np_iint_u_2d
            res_d_2d = profile_2d*np_iint_d_2d

            # 0.5 to have the same meaning for scale factor as in FullProf
            res_u_1d += 0.5*scale*res_u_2d.sum(axis=1)
            res_d_1d += 0.5*scale*res_d_2d.sum(axis=1)

            if flag_internal:
                peak.numpy_to_items()
            l_peak.append(peak)
            l_dd_out.append(dd_out)

        proc.numpy_ttheta_corrected = tth_zs
        proc.numpy_intensity_up_net = res_u_1d
        proc.numpy_intensity_down_net = res_d_1d
        proc.numpy_intensity_net = res_u_1d+res_d_1d
        proc.numpy_intensity_diff_total = res_u_1d-res_d_1d
        if flag_polarized:
            proc.numpy_intensity_up_total = res_u_1d+int_bkgd
            proc.numpy_intensity_down_total = res_d_1d+int_bkgd
            proc.numpy_intensity_total = res_u_1d+res_d_1d+int_bkgd+int_bkgd
        else:
            proc.numpy_intensity_up_total = res_u_1d+0.5*int_bkgd
            proc.numpy_intensity_down_total = res_d_1d+0.5*int_bkgd
            proc.numpy_intensity_total = res_u_1d+res_d_1d+int_bkgd

        if flag_internal:
            proc.numpy_to_items()
            l_calc_objs = l_refln + l_refln_s + l_peak
            l_calc_objs.append(proc)
            self.add_items(l_calc_objs)

        return proc, l_peak, l_refln, l_dd_out

    def calc_chi_sq(self, l_crystal, flag_internal=True):
        """
        Calculate chi square.

        Arguments
        ---------
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate or to use internal objects.
                   It should be True if user call the function.
                   It's True by default.

        Output
        ------
            - chi_sq_val: chi square of flip ratio
              (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
            - n: number of measured reflections
        """
        meas = self.pd_meas
        flag_polarized = meas.is_polarized()

        tth = meas.numpy_ttheta
        if flag_polarized:
            int_u_exp = meas.numpy_intensity_up
            sint_u_exp = meas.numpy_intensity_up_sigma
            int_d_exp = meas.numpy_intensity_down
            sint_d_exp = meas.numpy_intensity_down_sigma
        else:
            int_exp = meas.numpy_intensity
            sint_exp = meas.numpy_intensity_sigma

        l_peak_in, l_refln_in = [], []
        l_refln_susceptibility_in = []
        l_dd_in = []
        flag_1 = not(flag_internal)
        flag_2 = (self.dd is not None)
        if (flag_1 & flag_2):
            for phase_item in self.phase.items:
                crystal = None
                for cryst in l_crystal:
                    if cryst.data_name.lower() == phase_item.label.lower():
                        crystal = cryst
                        break
                attr_peak = f"pd_peak_{crystal.data_name:}"
                attr_refln = f"refln_{crystal.data_name:}"
                attr_refln_s = f"refln_susceptibility_{crystal.data_name:}"
                l_peak_in.append(getattr(self, attr_peak))
                l_refln_in.append(getattr(self, attr_refln))
                l_refln_susceptibility_in.append(getattr(self, attr_refln_s))

        range_ = self.range
        cond_in = numpy.ones(tth.shape, dtype=bool)
        if range_ is not None:
            range_min = numpy.array(range_.ttheta_min, dtype=float)
            range_max = numpy.array(range_.ttheta_max, dtype=float)

            cond_in = numpy.logical_and(cond_in, tth >= range_min)
            cond_in = numpy.logical_and(cond_in, tth <= range_max)

        tth_in = tth[cond_in]
        if flag_polarized:
            int_u_exp_in = int_u_exp[cond_in]
            sint_u_exp_in = sint_u_exp[cond_in]
            int_d_exp_in = int_d_exp[cond_in]
            sint_d_exp_in = sint_d_exp[cond_in]
        else:
            int_exp_in = int_exp[cond_in]
            sint_exp_in = sint_exp[cond_in]

        proc, l_peak, l_refln, l_dd_out = self.calc_profile(
            tth_in, l_crystal, l_peak_in=l_peak_in, l_refln_in=l_refln_in,
            l_refln_susceptibility_in=l_refln_susceptibility_in,
            l_dd_in=l_dd_in, flag_internal=flag_internal,
            flag_polarized=flag_polarized)

        self.dd = l_dd_out

        if flag_polarized:
            proc.numpy_intensity_up = int_u_exp_in
            proc.numpy_intensity_up_sigma = sint_u_exp_in
            proc.numpy_intensity_down = int_d_exp_in
            proc.numpy_intensity_down_sigma = sint_d_exp_in
            proc.numpy_intensity = int_u_exp_in+int_d_exp_in
            proc.numpy_intensity_sigma = numpy.sqrt(
                numpy.square(sint_u_exp_in) + numpy.square(sint_d_exp_in))
        else:
            proc.numpy_intensity = int_exp_in
            proc.numpy_intensity_sigma = sint_exp_in

        int_u_mod = proc.numpy_intensity_up_total
        int_d_mod = proc.numpy_intensity_down_total

        if flag_polarized:
            sint_sum_exp_in = (sint_u_exp_in**2 + sint_d_exp_in**2)**0.5
            chi_sq_u = ((int_u_mod-int_u_exp_in)/sint_u_exp_in)**2
            chi_sq_d = ((int_d_mod-int_d_exp_in)/sint_d_exp_in)**2
            chi_sq_sum = ((int_u_mod+int_d_mod-int_u_exp_in-int_d_exp_in) /
                          sint_sum_exp_in)**2
            chi_sq_dif = ((int_u_mod-int_d_mod-int_u_exp_in+int_d_exp_in) /
                          sint_sum_exp_in)**2

            cond_u = numpy.logical_not(numpy.isnan(chi_sq_u))
            cond_d = numpy.logical_not(numpy.isnan(chi_sq_d))
            cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))
            cond_dif = numpy.logical_not(numpy.isnan(chi_sq_dif))
        else:
            chi_sq_sum = ((int_u_mod+int_d_mod-int_exp_in)/sint_exp_in)**2
            cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))

        # exclude region
        exclude = self.exclude
        if exclude is not None:
            l_excl_tth_min = exclude.ttheta_min
            l_excl_tth_max = exclude.ttheta_max
            if flag_polarized:
                for excl_tth_min, excl_tth_max in zip(l_excl_tth_min,
                                                      l_excl_tth_max):
                    cond_1 = numpy.logical_or(tth_in < 1.*excl_tth_min,
                                              tth_in > 1.*excl_tth_max)
                    cond_u = numpy.logical_and(cond_u, cond_1)
                    cond_d = numpy.logical_and(cond_d, cond_1)
                    cond_sum = numpy.logical_and(cond_sum, cond_1)
            else:
                for excl_tth_min, excl_tth_max in zip(l_excl_tth_min,
                                                      l_excl_tth_max):
                    cond_1 = numpy.logical_or(tth_in < 1.*excl_tth_min,
                                              tth_in > 1.*excl_tth_max)
                    cond_sum = numpy.logical_and(cond_sum, cond_1)

        chi_sq_sum_val = (chi_sq_sum[cond_sum]).sum()
        n_sum = cond_sum.sum()

        if flag_polarized:

            chi_sq_u_val = (chi_sq_u[cond_u]).sum()
            n_u = cond_u.sum()

            chi_sq_d_val = (chi_sq_d[cond_d]).sum()
            n_d = cond_d.sum()

            chi_sq_dif_val = (chi_sq_dif[cond_dif]).sum()
            n_dif = cond_dif.sum()

            chi2 = self.chi2

            flag_u = chi2.up
            flag_d = chi2.down
            flag_sum = chi2.sum
            flag_dif = chi2.diff

            chi_sq_val = (int(flag_u)*chi_sq_u_val + int(flag_d)*chi_sq_d_val +
                          int(flag_sum)*chi_sq_sum_val +
                          int(flag_dif)*chi_sq_dif_val)
            n = (int(flag_u)*n_u + int(flag_d)*n_d + int(flag_sum)*n_sum +
                 int(flag_dif)*n_dif)
        else:
            chi_sq_val = chi_sq_sum_val
            n = n_sum
        # d_exp_out = {"chi_sq_val": chi_sq_val, "n": n}
        # d_exp_out.update(d_exp_prof_out)
        # print(f"chi_sq_val/n: {chi_sq_val/n:.2f} , \
        #         chi_sq_val: {chi_sq_val:.2f}")

        if flag_internal:
            refine_ls = RefineLs(number_reflns=n,
                                 goodness_of_fit_all=chi_sq_val/float(n),
                                 weighting_scheme="sigma")
            self.refine_ls = refine_ls
            proc.numpy_to_items()
        return chi_sq_val, n

    def simulation(self, l_crystal, ttheta_start: float = 4.,
                   ttheta_end: float = 120., ttheta_step: float = 0.1,
                   flag_polarized: bool = True) -> NoReturn:
        """Simulate."""
        n_points = int(round((ttheta_end-ttheta_start)/ttheta_step))
        tth_in = numpy.linspace(ttheta_start, ttheta_end, n_points)
        if isinstance(l_crystal, Crystal):
            l_crystal = [l_crystal]
        self.calc_profile(tth_in, l_crystal, l_peak_in=None, l_refln_in=None,
                          l_refln_susceptibility_in=None, l_dd_in=None,
                          flag_internal=True, flag_polarized=flag_polarized)
        return

    def calc_iint(self, index_h, index_k, index_l, crystal: Crystal,
                  flag_internal: bool = True):
        """
        Calculate the integrated intensity for h, k, l reflections.

        Keyword Arguments
        -----------------
            - h, k, l: 1D numpy array of Miller indexes, dtype = int32
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate or to use internal objects.
                   It should be True if user call the function.
                   It's True by default.

        Output Arguments
        ----------------
            - iint_u: 1D numpy array of integrated intensity up, dtype = float
            - iint_d: 1D numpy array of integrated intensity up, dtype = float
            - refln: ReflnL object of cryspy library (nuclear structure factor)
            - refln_s: ReflnSusceptibilityL object of cryspy library
              (nuclear structure factor)
        """
        field = self.setup.field
        if field is None:
            field = 0.

        diffrn_radiation = self.diffrn_radiation
        if diffrn_radiation is not None:
            p_u = float(diffrn_radiation.polarization)
            p_d = (2.*float(diffrn_radiation.efficiency)-1)*p_u
        else:
            p_u = 0.0
            p_d = 0.0

        refln = crystal.calc_refln(index_h, index_k, index_l,
                                   flag_internal=flag_internal)
        f_nucl = refln.numpy_f_calc

        refln_s = crystal.calc_refln_susceptibility(
            index_h, index_k, index_l, flag_internal=flag_internal)

        sft_11 = refln_s.numpy_chi_11_calc
        sft_12 = refln_s.numpy_chi_12_calc
        sft_13 = refln_s.numpy_chi_13_calc
        sft_21 = refln_s.numpy_chi_21_calc
        sft_22 = refln_s.numpy_chi_22_calc
        sft_23 = refln_s.numpy_chi_23_calc
        sft_31 = refln_s.numpy_chi_31_calc
        sft_32 = refln_s.numpy_chi_32_calc
        sft_33 = refln_s.numpy_chi_33_calc

        sftm_11 = refln_s.numpy_moment_11_calc
        sftm_12 = refln_s.numpy_moment_12_calc
        sftm_13 = refln_s.numpy_moment_13_calc
        sftm_21 = refln_s.numpy_moment_21_calc
        sftm_22 = refln_s.numpy_moment_22_calc
        sftm_23 = refln_s.numpy_moment_23_calc
        sftm_31 = refln_s.numpy_moment_31_calc
        sftm_32 = refln_s.numpy_moment_32_calc
        sftm_33 = refln_s.numpy_moment_33_calc

        _11, _12 = sftm_11+field*sft_11, sftm_12+field*sft_12
        _21, _13 = sftm_21+field*sft_21, sftm_13+field*sft_13
        _22, _23 = sftm_22+field*sft_22, sftm_23+field*sft_23
        _31, _32 = sftm_31+field*sft_31, sftm_32+field*sft_32
        _33 = sftm_33+field*sft_33
        _ij = (_11, _12, _13, _21, _22, _23, _31, _32, _33)
        cell = crystal.cell

        # k_loc = cell.calc_k_loc(h, k, l)
        t_ij = cell.calc_m_t(index_h, index_k, index_l)
        # FIXME: I would like to recheck the expression for T
        #        and expression SIGMA = T^T CHI T
        t_tr_ij = (t_ij[0], t_ij[3], t_ij[6],
                   t_ij[1], t_ij[4], t_ij[7],
                   t_ij[2], t_ij[5], t_ij[8])
        th_11, th_12, th_13, th_21, th_22, th_23, th_31, th_32, th_33 = \
            calc_mRmCmRT(t_tr_ij, _ij)

        # fm_p_sq = (field**2)*abs(0.5*(th_11*th_11.conjugate()+
        #            th_22*th_22.conjugate())+th_12*th_12.conjugate())
        # fm_p_field = field*0.5*(th_11+th_22)
        fm_p_sq = abs(0.5 * (th_11 * th_11.conjugate() +
                             th_22 * th_22.conjugate()) +
                      th_12 * th_12.conjugate())
        fm_p_field = 0.5*(th_11 + th_22)
        cross = 2.*(f_nucl.real*fm_p_field.real+f_nucl.imag*fm_p_field.imag)

        iint_u = abs(f_nucl*f_nucl.conjugate()) + fm_p_sq + p_u*cross
        iint_d = abs(f_nucl*f_nucl.conjugate()) + fm_p_sq - p_d*cross

        # d_info_out = {"iint_u": iint_u, "iint_d": iint_d}
        # d_info_out.update(d_info_cry)
        return iint_u, iint_d, refln, refln_s

    def _gauss_pd(self, tth_2d):
        """One dimensional gauss powder diffraction."""
        ag, bg = self.ag, self.bg
        val_1 = bg*tth_2d**2
        val_2 = numpy.where(val_1 < 5., numpy.exp(-val_1), 0.)
        self.gauss_pd = ag*val_2

    def _lor_pd(self, tth_2d):
        """One dimensional lorentz powder diffraction."""
        al, bl = self.al, self.bl
        self.lor_pd = al*1./(1.+bl*tth_2d**2)

    def calc_shape_profile(self, tth, tth_hkl, i_g=0.):
        """
        Calculate shape profile.

        Calculate profile in the range ttheta for reflections placed on
        ttheta_hkl with i_g parameter by default equal to zero

        tth, tth_hkl in degrees
        """
        zero_shift = float(self.setup.offset_ttheta)
        tth_zs = tth-zero_shift

        resolution = self.pd_instr_resolution

        h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = resolution.calc_resolution(
            tth_hkl, i_g=i_g)

        tth_2d, tth_hkl_2d = numpy.meshgrid(tth_zs, tth_hkl, indexing="ij")

        self.ag = numpy.meshgrid(tth_zs, a_g, indexing="ij")[1]
        self.bg = numpy.meshgrid(tth_zs, b_g, indexing="ij")[1]
        self.al = numpy.meshgrid(tth_zs, a_l, indexing="ij")[1]
        self.bl = numpy.meshgrid(tth_zs, b_l, indexing="ij")[1]
        eta_2d = numpy.meshgrid(tth_zs, eta, indexing="ij")[1]
        self.eta = eta_2d

        self._gauss_pd(tth_2d-tth_hkl_2d)
        self._lor_pd(tth_2d-tth_hkl_2d)
        g_pd_2d = self.gauss_pd
        l_pd_2d = self.lor_pd

        np_shape_2d = eta_2d * l_pd_2d + (1.-eta_2d) * g_pd_2d

        asymmetry = self.asymmetry
        if asymmetry is not None:
            np_ass_2d = asymmetry.calc_asymmetry(tth_zs, tth_hkl, h_pv)
        else:
            np_ass_2d = numpy.ones(shape=np_shape_2d.shape)

        # Lorentz factor
        tth_rad = tth_zs*numpy.pi/180.
        np_lor_1d = 1./(numpy.sin(tth_rad)*numpy.sin(0.5*tth_rad))

        np_lor_2d = numpy.meshgrid(np_lor_1d, tth_hkl, indexing="ij")[0]

        profile_2d = np_shape_2d*np_ass_2d*np_lor_2d

        return profile_2d, tth_zs, h_pv

    def params_to_cif(self, separator="_", flag: bool = False,
                      flag_minimal: bool = True) -> str:
        """Save parameters to cif format."""
        ls_out = []
        l_cls = (Setup, PdInstrResolution, DiffrnRadiation, Chi2, Range,
                 Extinction, PdInstrReflexAsymmetry, Texture, PhaseL, ExcludeL,
                 PdBackgroundL)
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def data_to_cif(self, separator="_", flag: bool = False,
                    flag_minimal: bool = True) -> str:
        """Save data to cif format."""
        ls_out = []
        l_cls = (PdMeasL, )
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def calc_to_cif(self, separator="_", flag=False, flag_minimal=True) -> str:
        """Save calculations to cif format."""
        ls_out = []
        l_cls = (RefineLs, ReflnL, ReflnSusceptibilityL, PdPeakL, PdProcL)
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def apply_constraints(self):
        """Apply constraints."""
        if self.pd_meas is not None:
            self.pd_meas.apply_constraints()

# s_cont = """
#  data_pnd
#  _setup_wavelength     0.840
#  _setup_field          1.000
#  _setup_offset_2theta -0.385(12)
#  _chi2_sum  True
#  _chi2_diff False
#  _chi2_up   False
#  _chi2_down False
#  _range_2theta_min     4.000
#  _range_2theta_max    80.000
#  loop_
#  _pd_background_2theta
#  _pd_background_intensity
#   4.5 256.0
#  40.0 158.0
#  80.0  65.0
#  loop_
#  _exclude_2theta_min
#  _exclude_2theta_max
#  0.0 1.0
#  _pd_instr_reflex_asymmetry_p1 0.0
#  _pd_instr_reflex_asymmetry_p2 0.0
#  _pd_instr_reflex_asymmetry_p3 0.0
#  _pd_instr_reflex_asymmetry_p4 0.0
#  _diffrn_radiation_polarization 0.0
#  _diffrn_radiation_efficiency   1.0
#  _pd_instr_resolution_u 16.9776
#  _pd_instr_resolution_v -2.8357
#  _pd_instr_resolution_w  0.5763
#  _pd_instr_resolution_x  0.0
#  _pd_instr_resolution_y  0.0
#  loop_
#  _phase_label
#  _phase_scale
#  _phase_igsize
#  Fe3O4 0.02381 0.0
#  loop_
#  _pd_meas_2theta
#  _pd_meas_intensity_up
#  _pd_meas_intensity_up_sigma
#  _pd_meas_intensity_down
#  _pd_meas_intensity_down_sigma
#  4.0 465.80 128.97 301.88 129.30
#  4.2 323.78 118.22 206.06 120.00
#  4.4 307.14 115.90 230.47 116.53
#  loop_
#  _pd_peak_index_h
#  _pd_peak_index_k
#  _pd_peak_index_l
#  _pd_peak_mult
#  _pd_peak_ttheta
#  _pd_peak_intensity_up
#  _pd_peak_intensity_down
#  _pd_peak_width_2theta
#  1 1 1  8  9.748 128.15576 128.15576 0.677
#  2 0 0  6 11.260   0.00000   0.00000 0.680
#  2 2 0 12 15.950  94.21107  94.21107 0.716
#  loop_
#  _pd_proc_2theta
#  _pd_proc_2theta_corrected
#  _pd_proc_intensity_up_net
#  _pd_proc_intensity_down_net
#  _pd_proc_intensity_up_total
#  _pd_proc_intensity_down_total
#  _pd_proc_intensity_bkg_calc
#  _pd_proc_intensity_up
#  _pd_proc_intensity_up_sigma
#  _pd_proc_intensity_down
#  _pd_proc_intensity_down_sigma
#  4.000 4.385 0.0 0.0 256.0 256.0 256.0 465.8 128.97000 301.88000 129.30000
#  4.200 4.585 0.0 0.0 256.0 256.0 256.0 323.78 118.22000 206.06000 120.00000
#  4.400 4.785 0.0 0.0 256.0 256.0 256.0 307.14 115.90000 230.47000 116.53000
# """
# obj = Pd.from_cif(s_cont)
# print(obj)
# for var_name in obj.get_variable_names():
#     print(var_name)



import numpy
from typing import NoReturn
from cryspy.A_functions_base.function_1_gamma_nu import \
    recal_int_to_gammanu_grid
from cryspy.B_parent_classes.cl_1_item import ItemN

class Pd2dMeas(ItemN):
    """
    This section contains the measured diffractogram and information
    about the conditions used for the measurement of the diffraction 
    data set, prior to processing and application of correction
    terms.

    Mandatory attributes:
        - ub_11, ub_12, ub_13, ub_21, ub_22, ub_23, ub_31, ub_32, ub_33

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """
    ATTR_MANDATORY_NAMES = (
        "ttheta_phi_intensity_up", "ttheta_phi_intensity_up_sigma",
        "ttheta_phi_intensity_down", "ttheta_phi_intensity_down_sigma")
    ATTR_MANDATORY_TYPES = (str, str, str, str)
    # ("matrix", "matrix", "matrix", "matrix")
    ATTR_MANDATORY_CIF = (
        "2theta_phi_intensity_up", "2theta_phi_intensity_up_sigma",
        "2theta_phi_intensity_down", "2theta_phi_intensity_down_sigma")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("ttheta", "phi", "intensity_up", "intensity_up_sigma",
                      "intensity_down", "intensity_down_sigma")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "pd2d_meas"

    def __init__(self, **kwargs) -> NoReturn:
        super(Pd2dMeas, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        """Redefine form_object."""
        flag = (self.is_attribute("ttheta_phi_intensity_up") &
                self.is_attribute("ttheta_phi_intensity_up_sigma") &
                self.is_attribute("ttheta_phi_intensity_down") &
                self.is_attribute("ttheta_phi_intensity_down_sigma"))
        if not(flag):
            return
        l_1 = (self.ttheta_phi_intensity_up).strip().split("\n")
        l_2 = (self.ttheta_phi_intensity_up_sigma).strip().split("\n")
        l_3 = (self.ttheta_phi_intensity_down).strip().split("\n")
        l_4 = (self.ttheta_phi_intensity_down_sigma).strip().split("\n")

        l_ttheta = numpy.array([_ for _ in l_1[0].strip().split()[1:]],
                               dtype=float)
        l_phi, ll_intensity_up, ll_intensity_up_sigma = [], [], []
        ll_intensity_down, ll_intensity_down_sigma = [], []
        for line_1, line_2, line_3, line_4 in zip(l_1[1:], l_2[1:], l_3[1:],
                                                  l_4[1:]):
            _l_1 = line_1.strip().split()
            _l_2 = line_2.strip().split()
            _l_3 = line_3.strip().split()
            _l_4 = line_4.strip().split()
            l_phi.append(float(_l_1[0]))
            ll_intensity_up.append(_l_1[1:])
            ll_intensity_up_sigma.append(_l_2[1:])
            ll_intensity_down.append(_l_3[1:])
            ll_intensity_down_sigma.append(_l_4[1:])

        ll_intensity_up = numpy.array(ll_intensity_up, dtype=float).transpose()
        ll_intensity_up_sigma = numpy.array(ll_intensity_up_sigma, dtype=float).transpose()
        ll_intensity_down = numpy.array(ll_intensity_down, dtype=float).transpose()
        ll_intensity_down_sigma = numpy.array(ll_intensity_down_sigma, dtype=float).transpose()

        self.__dict__["ttheta"] = l_ttheta
        self.__dict__["phi"] = numpy.array(l_phi, dtype=float)
        self.__dict__["intensity_up"] = ll_intensity_up
        self.__dict__["intensity_up_sigma"] = ll_intensity_up_sigma
        self.__dict__["intensity_down"] = ll_intensity_down
        self.__dict__["intensity_down_sigma"] = ll_intensity_down_sigma


    def form_ttheta_phi_intensity_up(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_up")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_up
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_up"] = "\n".join(ls_out)

    def form_ttheta_phi_intensity_up_sigma(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_up_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_up_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_up_sigma"] = "\n".join(ls_out)

    def form_ttheta_phi_intensity_down(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_down")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_down
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_down"] = "\n".join(ls_out)

    def form_ttheta_phi_intensity_down_sigma(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_down_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_down_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_down_sigma"] = "\n".join(ls_out)

    def recalc_to_gamma_nu_grid(self):
        l_tth_grid = numpy.array(self.ttheta)*numpy.pi/180.
        l_phi_grid = numpy.array(self.phi)*numpy.pi/180.
        int_u = numpy.array(self.intensity_up, dtype=float).transpose()
        int_d = numpy.array(self.intensity_down, dtype=float).transpose()
        int_sum = int_u + int_d
        int_diff = int_u - int_d

        min_tth, max_tth = min(l_tth_grid), max(l_tth_grid)
        min_phi, max_phi = min(l_phi_grid), max(l_phi_grid)

        min_gamma, max_gamma = min_tth, max_tth
        num_gamma = len(l_tth_grid)

        min_nu, max_nu = -10.*numpy.pi/180., 15.*numpy.pi/180.
        num_nu = len(l_phi_grid)

        l_gamma_grid = numpy.linspace(min_gamma, max_gamma, num=num_gamma)
        l_nu_grid = numpy.linspace(min_nu, max_nu, num=num_nu)

        int_u_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_u, l_gamma_grid, l_nu_grid), dtype=float)
        int_d_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_d, l_gamma_grid, l_nu_grid), dtype=float)
        int_sum_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_sum, l_gamma_grid, l_nu_grid), dtype=float)
        int_diff_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_diff, l_gamma_grid, l_nu_grid), dtype=float)

        return l_gamma_grid*180./numpy.pi, l_nu_grid*180./numpy.pi, [int_u_out, int_d_out, int_sum_out, int_diff_out]
        

# s_cont = """

#  _pd2d_meas_2theta_phi_intensity_up
#  ;
#       2    4.5     40.0     80.0
#  -3.000 -350.0   -350.0   -400.0
#  41.000 -351.0   -350.0   -400.0
#  ;

#  _pd2d_meas_2theta_phi_intensity_up_sigma
#  ;
#       2    4.5     40.0     80.0
#  -3.000 -352.0   -350.0   -400.0
#  41.000 -353.0   -350.0   -400.0
#  ;

#  _pd2d_meas_2theta_phi_intensity_down
#  ;
#       2    4.5     40.0     80.0
#  -3.000 -354.0   -350.0   -400.0
#  41.000 -355.0   -350.0   -400.0
#  ;

#  _pd2d_meas_2theta_phi_intensity_down_sigma
#  ;
#       2    4.5     40.0     80.0
#  -3.000 -356.0   -350.0   -400.0
#  41.000 -357.0   -350.0   -400.0
#  ; 
# """

# obj = Pd2dMeas.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.ttheta, end="\n\n")
# print(obj.phi, end="\n\n")
# print(obj.intensity_up, end="\n\n")
# print(obj.intensity_up_sigma, end="\n\n")



"""Description of Diffrn class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"

from warnings import warn
from typing import NoReturn, List
import numpy
import scipy
import scipy.misc

from cryspy.A_functions_base.function_1_matrices import scalar_product
from cryspy.A_functions_base.function_1_roots import calc_roots

from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN

from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import \
    DiffrnOrientMatrix
from cryspy.C_item_loop_classes.cl_1_diffrn_refln import DiffrnReflnL
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import Phase
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs

from cryspy.D_functions_item_loop.function_1_flip_ratio import \
    calc_flip_ratio, calc_fm_perp_loc, calc_e_up_loc

from cryspy.E_data_classes.cl_1_crystal import Crystal


class Diffrn(DataN):
    """
    Diffrn class.

    Data items in the DIFFRN category record details about
    single diffraction measurements.

    Methods
    -------
        - calc_iint_u_d_flip_ratio
        - calc_fr
        - calc_fm_perp_loc
        - calc_chi_sq
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - estimate_FM


    Attributes
    ----------
        - setup (mandatory)
        - diffrn_radiation (mandatory)
        - diffrn_orient_matrix (mandatory)
        - diffrn_refln (mandatory)
        - extinction
        - phase
        - refln_#phase_name
        - refln_susceptibility_#phase_name
        - refine_ls
    """

    CLASSES_MANDATORY = (Setup, DiffrnRadiation, DiffrnOrientMatrix,
                         DiffrnReflnL)
    CLASSES_OPTIONAL = (Extinction, Phase, ReflnL, ReflnSusceptibilityL,
                        RefineLs)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "diffrn"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Diffrn, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_iint_u_d_flip_ratio(self, index_h, index_k, index_l, l_crystal,
                                 flag_internal: bool = True):
        """
        Calculate intensity for the given diffraction angle.

        Keyword Arguments
        -----------------
            h, k, l: 1D numpy array of Miller indexes
            l_crystal: list of crystal structures

        Output
        ------
            iint_u: intensity up (without scale factor)
            iint_d: intensity down (without scale factor)
            flip_ratio: flip ratio (intensity_up/intensity_down)

        """
        l_label = [_.data_name for _ in l_crystal]
        if self.phase.label in l_label:
            ind = l_label.index(self.phase.label)
            crystal = l_crystal[ind]
        elif self.data_name in l_label:
            self.phase = Phase(label=self.data_name)
            ind = l_label.index(self.data_name)
            crystal = l_crystal[ind]
        else:
            warn("Crystal not found. The first one is taken.")
            crystal = l_crystal[0]
            self.phase.label = crystal.label

        cell = crystal.cell
        volume_unit_cell = cell.volume
        sthovl = cell.calc_sthovl(index_h, index_k, index_l)
        k_loc_i = cell.calc_k_loc(index_h, index_k, index_l)

        setup = self.setup
        wavelength = setup.wavelength
        field_norm = setup.field
        try:
            ratio_lambdaover2 = setup.ratio_lambdaover2
            flag_lambdaover2 = True
        except AttributeError:
            ratio_lambdaover2 = None
            flag_lambdaover2 = False

        diffrn_orient_matrix = self.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        e_up_loc = calc_e_up_loc(0., 0., 0., u_matrix)

        diffrn_radiation = self.diffrn_radiation
        polarization = diffrn_radiation.polarization
        flipper_efficiency = diffrn_radiation.efficiency

        if self.is_attribute("extinction"):
            extinction = self.extinction
            model_extinction = extinction.model
            radius = extinction.radius
            mosaicity = extinction.mosaicity
        else:
            model_extinction = ""
            radius = None
            mosaicity = None

        if flag_lambdaover2:
            index_2h = 2 * index_h
            index_2k = 2 * index_k
            index_2l = 2 * index_l

        if flag_internal:
            refln = crystal.calc_refln(index_h, index_k, index_l)
            f_nucl = numpy.array(refln.f_calc, dtype=complex)
        else:
            f_nucl = crystal.calc_f_nucl(index_h, index_k, index_l)

        if flag_lambdaover2:
            f_nucl_2hkl = crystal.calc_f_nucl(index_2h, index_2k, index_2l)
        else:
            f_nucl_2hkl = None

        if flag_internal:
            refln_s = crystal.calc_refln_susceptibility(index_h, index_k,
                                                        index_l)
            sft_ij = (refln_s.numpy_chi_11_calc,
                      refln_s.numpy_chi_12_calc,
                      refln_s.numpy_chi_13_calc,
                      refln_s.numpy_chi_21_calc,
                      refln_s.numpy_chi_22_calc,
                      refln_s.numpy_chi_23_calc,
                      refln_s.numpy_chi_31_calc,
                      refln_s.numpy_chi_32_calc,
                      refln_s.numpy_chi_33_calc)

            sftm_ij = (refln_s.numpy_moment_11_calc,
                       refln_s.numpy_moment_12_calc,
                       refln_s.numpy_moment_13_calc,
                       refln_s.numpy_moment_21_calc,
                       refln_s.numpy_moment_22_calc,
                       refln_s.numpy_moment_23_calc,
                       refln_s.numpy_moment_31_calc,
                       refln_s.numpy_moment_32_calc,
                       refln_s.numpy_moment_33_calc)
        else:
            chi_m = crystal.calc_susceptibility_moment_tensor(
                index_h, index_k, index_l)
            sft_ij = chi_m[:9]
            sftm_ij = chi_m[9:]

        fm_perp_loc = calc_fm_perp_loc(e_up_loc, field_norm, k_loc_i, sft_ij,
                                       sftm_ij)

        if flag_lambdaover2:
            chi_m_2hkl = crystal.calc_susceptibility_moment_tensor(
                index_2h, index_2k, index_2l)
            sft_ij_2hkl = chi_m_2hkl[:9]
            sftm_ij_2hkl = chi_m_2hkl[9:]
            fm_perp_loc_2hkl = calc_fm_perp_loc(e_up_loc, field_norm, k_loc_i,
                                                sft_ij_2hkl, sftm_ij_2hkl)
        else:
            fm_perp_loc_2hkl = None

        flip_ratio, dder = calc_flip_ratio(
            sthovl, wavelength, field_norm, u_matrix, polarization,
            flipper_efficiency, volume_unit_cell, model_extinction, radius,
            mosaicity, f_nucl, fm_perp_loc,
            ratio_lambdaover2=ratio_lambdaover2, f_nucl_2hkl=f_nucl_2hkl,
            fm_perp_loc_2hkl=fm_perp_loc_2hkl)

        if flag_internal:
            refln.loop_name = crystal.data_name
            refln_s.loop_name = crystal.data_name
            self.add_items([refln, refln_s])

        return flip_ratio, dder

    # FIXME: merge with calc_iint_u_d_flip_ratio
    def calc_fr(self, cell, f_nucl, f_m_perp, delta_f_nucl=None,
                delta_f_m_perp=None):
        """
        Calculate flip ratio.

        By given nuclear structure factor and
        perpendicular component of magnetic structure factor
        (in 10**-12 cm)
        cell (only for extinction correction)
        f_nucl [hkl]
        f_m_perp [hkl]
        delta_f_nucl [hkl, parameters]
        delta_f_m_perp [hkl, parameters]
        """
        f_n_sq = (f_nucl * f_nucl.conjugate()).real
        f_m_p_x, f_m_p_y, f_m_p_z = f_m_perp
        delta_f_m_p_x, delta_f_m_p_y, delta_f_m_p_z = delta_f_m_perp

        f_m_p_sq = (f_m_p_x*f_m_p_x.conjugate() +
                    f_m_p_y*f_m_p_y.conjugate() +
                    f_m_p_z*f_m_p_z.conjugate()).real

        diffrn_radiation = self.diffrn_radiation
        diffrn_orient_matrix = self.diffrn_orient_matrix
        e_up = diffrn_orient_matrix.calc_e_up()
        p_u = float(diffrn_radiation.polarization)
        p_d = (2. * float(diffrn_radiation.efficiency) - 1.) * p_u

        f_m_p_vert = scalar_product(f_m_perp, e_up)

        if delta_f_m_perp is not None:
            delta_f_m_p_vert = scalar_product(delta_f_m_perp, e_up)

        two_f_nucl_f_m_p_vert = (f_nucl * f_m_p_vert.conjugate() +
                                 f_m_p_vert * f_nucl.conjugate()).real

        delta_two_f_nucl_f_m_p_vert = (
            f_nucl[:, numpy.newaxis]*delta_f_m_p_vert.conjugate() +
            delta_f_m_p_vert*f_nucl.conjugate()[:, numpy.newaxis]).real

        extinction = self.extinction
        diffrn_refln = self.diffrn_refln
        index_h = diffrn_refln.numpy_index_h
        index_k = diffrn_refln.numpy_index_k
        index_l = diffrn_refln.numpy_index_l

        setup = self.setup
        wavelength = setup.wavelength

        delta_f_m_p_sq = 2.*(
         numpy.real(f_m_p_x)[:, numpy.newaxis] * numpy.real(delta_f_m_p_x) +
         numpy.imag(f_m_p_x)[:, numpy.newaxis] * numpy.imag(delta_f_m_p_x) +
         numpy.real(f_m_p_y)[:, numpy.newaxis] * numpy.real(delta_f_m_p_y) +
         numpy.imag(f_m_p_y)[:, numpy.newaxis] * numpy.imag(delta_f_m_p_y) +
         numpy.real(f_m_p_z)[:, numpy.newaxis] * numpy.real(delta_f_m_p_z) +
         numpy.imag(f_m_p_z)[:, numpy.newaxis] * numpy.imag(delta_f_m_p_z))

        # extinction correction
        f_m_p_vert_sq = (f_m_p_vert * f_m_p_vert.conjugate()).real
        delta_f_m_p_vert_sq = 2.*(
            numpy.real(f_m_p_vert)[:, numpy.newaxis] *
            numpy.real(delta_f_m_p_vert) +
            numpy.imag(f_m_p_vert)[:, numpy.newaxis] *
            numpy.imag(delta_f_m_p_vert))

        fnp = two_f_nucl_f_m_p_vert
        fp_sq = f_n_sq + f_m_p_sq + two_f_nucl_f_m_p_vert
        fm_sq = f_n_sq + f_m_p_sq - two_f_nucl_f_m_p_vert
        fpm_sq = f_m_p_sq - f_m_p_vert_sq

        delta_fnp = delta_two_f_nucl_f_m_p_vert
        delta_fp_sq = delta_f_m_p_sq + delta_two_f_nucl_f_m_p_vert
        delta_fm_sq = delta_f_m_p_sq - delta_two_f_nucl_f_m_p_vert
        delta_fpm_sq = delta_f_m_p_sq - delta_f_m_p_vert_sq

        if extinction is not None:
            yp, dderp = extinction.calc_extinction(
                cell, index_h, index_k, index_l,
                fp_sq, wavelength, flag_derivatives=True)
            delta_yp = dderp["der_yext__f_sq"]
            ym, dderm = extinction.calc_extinction(
                cell, index_h, index_k, index_l,
                fm_sq, wavelength, flag_derivatives=True)
            delta_ym = dderm["der_yext__f_sq"]
            ypm, dderpm = extinction.calc_extinction(
                cell, index_h, index_k, index_l,
                fpm_sq, wavelength, flag_derivatives=True)
            delta_ypm = dderpm["der_yext__f_sq"]
        else:
            yp = numpy.ones(shape=fp_sq.shape, dtype=float),
            delta_yp = numpy.zeros(shape=fp_sq.shape, dtype=float)
            ym = numpy.ones(shape=fp_sq.shape, dtype=float)
            delta_ym = numpy.zeros(shape=fp_sq.shape, dtype=float)
            ypm = numpy.ones(shape=fp_sq.shape, dtype=float)
            delta_ypm = numpy.zeros(shape=fp_sq.shape, dtype=float)

        pppl = 0.5 * ((1 + p_u) * yp + (1 - p_u) * ym)
        ppmin = 0.5 * ((1 - p_d) * yp + (1 + p_d) * ym)
        pmpl = 0.5 * ((1 + p_u) * yp - (1 - p_u) * ym)
        pmmin = 0.5 * ((1 - p_d) * yp - (1 + p_d) * ym)

        delta_pppl = 0.5 * ((1 + p_u) * delta_yp[:, numpy.newaxis] *
                            delta_fp_sq + (1 - p_u) *
                            delta_ym[:, numpy.newaxis] * delta_fm_sq)
        delta_ppmin = 0.5 * ((1 - p_d) * delta_yp[:, numpy.newaxis] *
                             delta_fp_sq + (1 + p_d) *
                             delta_ym[:, numpy.newaxis] * delta_fm_sq)
        delta_pmpl = 0.5 * ((1 + p_u) * delta_yp[:, numpy.newaxis] *
                            delta_fp_sq - (1 - p_u) *
                            delta_ym[:, numpy.newaxis] * delta_fm_sq)
        delta_pmmin = 0.5 * ((1 - p_d) * delta_yp[:, numpy.newaxis] *
                             delta_fp_sq - (1 + p_d) *
                             delta_ym[:, numpy.newaxis] * delta_fm_sq)

        iint_u = (f_n_sq+f_m_p_vert_sq)*pppl+pmpl*fnp+ypm*fpm_sq
        iint_d = (f_n_sq+f_m_p_vert_sq)*ppmin+pmmin*fnp+ypm*fpm_sq

        delta_iint_u = (((f_n_sq + f_m_p_vert_sq)[:, numpy.newaxis] *
                         delta_pppl + delta_pmpl * fnp[:, numpy.newaxis] +
                         (delta_ypm * fpm_sq)[:, numpy.newaxis] * delta_fpm_sq)
                        + (delta_f_m_p_vert_sq) * pppl[:, numpy.newaxis] +
                        pmpl[:, numpy.newaxis] * delta_fnp +
                        ypm[:, numpy.newaxis] * delta_fpm_sq)

        delta_iint_d = (((f_n_sq + f_m_p_vert_sq)[:, numpy.newaxis] *
                         delta_ppmin + delta_pmmin * fnp[:, numpy.newaxis] +
                         (delta_ypm * fpm_sq)[:, numpy.newaxis] * delta_fpm_sq)
                        + (delta_f_m_p_vert_sq) * ppmin[:, numpy.newaxis] +
                        pmmin[:, numpy.newaxis] * delta_fnp +
                        ypm[:, numpy.newaxis] * delta_fpm_sq)

        f_r_m = iint_u/iint_d

        delta_f_r_m = (iint_d[:, numpy.newaxis] * delta_iint_u -
                       iint_u[:, numpy.newaxis] * delta_iint_d) / \
            (iint_d**2)[:, numpy.newaxis]

        return f_r_m, delta_f_r_m

    def calc_chi_sq(self, l_crystal: List[Crystal], flag_internal=True):
        """
        Calculate chi square.

        Keyword Arguments
        -----------------
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate internal objects
              (default is True)

        Output arguments
        ----------------
            - chi_sq_val: chi square of flip ratio
              (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
            - n: number of measured reflections
        """
        diffrn_refln = self.diffrn_refln
        np_h = diffrn_refln.numpy_index_h
        np_k = diffrn_refln.numpy_index_k
        np_l = diffrn_refln.numpy_index_l
        fr_exp = diffrn_refln.numpy_fr
        fr_sigma = diffrn_refln.numpy_fr_sigma
        fr_mod, dder = self.calc_iint_u_d_flip_ratio(
            np_h, np_k, np_l, l_crystal, flag_internal=flag_internal)

        if flag_internal:
            diffrn_refln.numpy_fr_calc = fr_mod
            # diffrn_refln.numpy_intensity_up_calc = int_u_mod
            # diffrn_refln.numpy_intensity_down_calc = int_d_mod
            diffrn_refln.numpy_to_items()

        chi_sq = ((fr_mod-fr_exp)/fr_sigma)**2
        chi_sq_val = (chi_sq[numpy.logical_not(numpy.isnan(chi_sq))]).sum()
        n = numpy.logical_not(numpy.isnan(chi_sq)).sum()

        if flag_internal:
            refine_ls = RefineLs(number_reflns=n,
                                 goodness_of_fit_all=chi_sq_val/float(n),
                                 weighting_scheme="sigma")
            self.refine_ls = refine_ls
        return chi_sq_val, n

    def params_to_cif(self, separator="_") -> str:
        """Save parameters in cif format."""
        ls_out = []
        l_cls = (Setup, DiffrnRadiation, DiffrnOrientMatrix, Extinction, Phase)
        l_obj = [item for item in self.items if type(item) in l_cls]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_obj])
        return "\n".join(ls_out)

    def data_to_cif(self, separator="_") -> str:
        """Save data in cif format."""
        ls_out = []
        l_cls = (DiffrnReflnL, )
        l_obj = [item for item in self.items if type(item) in l_cls]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_obj])
        return "\n".join(ls_out)

    def calc_to_cif(self, separator="_", flag=False) -> str:
        """Save calculations in cif format."""
        ls_out = []
        l_cls = (RefineLs, ReflnL, ReflnSusceptibilityL)
        l_obj = [item for item in self.items if type(item) in l_cls]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_obj])
        return "\n".join(ls_out)

    def estimate_FM(self, crystal, maxAbsFM: float = 5,
                    deltaFM: float = 0.001, flag_lambdaover2_correction=False):
        """
        Estimate magnetic structure factor.

        The method calculates magnetic structure factors (only real part) from
        flipping ratios based on give crystal structure.
        The crystal structure should be centrosymmetrical
        """
        setup = self.setup
        wavelength = setup.wavelength
        field_norm = setup.field

        diffrn_refln = self.diffrn_refln
        index_h = diffrn_refln.numpy_index_h
        index_k = diffrn_refln.numpy_index_k
        index_l = diffrn_refln.numpy_index_l
        fr_exp = diffrn_refln.numpy_fr
        fr_sigma = diffrn_refln.numpy_fr_sigma

        diffrn_radiation = self.diffrn_radiation
        polarization = diffrn_radiation.polarization
        flipper_efficiency = diffrn_radiation.efficiency

        extinction = self.extinction
        if extinction is not None:
            model_extinction = extinction.model
            radius = extinction.radius
            mosaicity = extinction.mosaicity
        else:
            model_extinction = ""
            radius = None
            mosaicity = None

        diffrn_orient_matrix = self.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        e_up_loc = calc_e_up_loc(0., 0., 0., u_matrix)

        refln = crystal.calc_refln(index_h, index_k, index_l)
        f_nucl = numpy.array(refln.f_calc, dtype=complex)

        cell = crystal.cell
        volume_unit_cell = cell.volume
        sthovl = cell.calc_sthovl(index_h, index_k, index_l)
        k_loc_i = cell.calc_k_loc(index_h, index_k, index_l)
        k_loc_1, k_loc_2, k_loc_3 = k_loc_i

        if (setup.is_attribute("ratio_lambdaover2") &
                flag_lambdaover2_correction):
            ratio_lambdaover2 = setup.ratio_lambdaover2
            flag_lambdaover2 = True
        else:
            ratio_lambdaover2 = None
            flag_lambdaover2 = False

        if (flag_lambdaover2 & flag_lambdaover2_correction):
            index_2h = 2 * index_h
            index_2k = 2 * index_k
            index_2l = 2 * index_l
            f_nucl_2hkl = crystal.calc_f_nucl(index_2h, index_2k, index_2l)
        else:
            f_nucl_2hkl = None

        if (flag_lambdaover2 & flag_lambdaover2_correction):
            chi_m_2hkl = crystal.calc_susceptibility_moment_tensor(
                index_2h, index_2k, index_2l)
            sft_ij_2hkl = chi_m_2hkl[:9]
            sftm_ij_2hkl = chi_m_2hkl[9:]
            fm_perp_loc_2hkl = calc_fm_perp_loc(e_up_loc, field_norm, k_loc_i,
                                                sft_ij_2hkl, sftm_ij_2hkl)
        else:
            fm_perp_loc_2hkl = None


        def function_to_find_fm(f_mag):
            """Calc flip ratio.

            Parameters
            ----------
            f_mag : float
                in muB.

            Returns
            -------
            flip_ratio : float
                flip ratio.
            """
            mag_loc_1 = e_up_loc[0]*f_mag*0.2695
            mag_loc_2 = e_up_loc[1]*f_mag*0.2695
            mag_loc_3 = e_up_loc[2]*f_mag*0.2695

            # vector product k x (0, 0, mag_3) x k
            mag_p_1 = (k_loc_3*mag_loc_1 - k_loc_1*mag_loc_3)*k_loc_3 - \
                (k_loc_1*mag_loc_2 - k_loc_2*mag_loc_1)*k_loc_2
            mag_p_2 = (k_loc_1*mag_loc_2 - k_loc_2*mag_loc_1)*k_loc_1 - \
                (k_loc_2*mag_loc_3 - k_loc_3*mag_loc_2)*k_loc_3
            mag_p_3 = (k_loc_2*mag_loc_3 - k_loc_3*mag_loc_2)*k_loc_2 - \
                (k_loc_3*mag_loc_1 - k_loc_1*mag_loc_3)*k_loc_1

            fm_perp_loc = (mag_p_1, mag_p_2, mag_p_3)
            flip_ratio, dder = calc_flip_ratio(
                sthovl, wavelength, field_norm, u_matrix, polarization,
                flipper_efficiency, volume_unit_cell, model_extinction, radius,
                mosaicity, f_nucl, fm_perp_loc,
                ratio_lambdaover2=ratio_lambdaover2,
                f_nucl_2hkl=f_nucl_2hkl,
                fm_perp_loc_2hkl=fm_perp_loc_2hkl)
            return flip_ratio

        lFMans, lFMansMin = [], []
        lFMansMin_sigma = []
        for _ind in range(fr_exp.size):
            def f(x):
                return (function_to_find_fm(x)[_ind] - fr_exp[_ind])
            FMans = calc_roots(f, -1*maxAbsFM-deltaFM, maxAbsFM+deltaFM)
            lFMans.append(FMans)
            if len(FMans) == 0:
                lFMansMin.append(None)
                lFMansMin_sigma.append(None)
            else:
                FMans_min = min(FMans)
                lFMansMin.append(FMans_min)
                der1 = scipy.misc.derivative(f, FMans_min, 0.1, 1)
                f_mag_sigma = abs(fr_sigma[_ind] / der1)
                lFMansMin_sigma.append(f_mag_sigma)

        ls_out = [f"loop_{crystal.data_name:}"]
        ls_out.append("_estimate_index_h\n_estimate_index_k")
        ls_out.append("_estimate_index_l\n_estimate_F_M\n_estimate_F_M_sigma")
        for ind_h, ind_k, ind_l, f_mag, f_sig in \
                zip(index_h, index_k, index_l, lFMansMin, lFMansMin_sigma):
            ls_out.append(f"{ind_h:} {ind_k:} {ind_l:} {f_mag:}  {f_sig:}")

        res = LoopN.from_cif("\n".join(ls_out))

        return res

    def apply_constraints(self):
        """Apply constraints."""
        pass

# s_cont = """
#  data_mono
#  _setup_wavelength     0.840
#  _setup_field          1.000

#  _diffrn_radiation_polarization 1.0
#  _diffrn_radiation_efficiency   1.0

#  _extinction_mosaicity 100.0
#  _extinction_radius    50.0
#  _extinction_model     gauss

#  _diffrn_orient_matrix_UB_11 6.59783
#  _diffrn_orient_matrix_UB_12 -6.99807
#  _diffrn_orient_matrix_UB_13 3.3663
#  _diffrn_orient_matrix_UB_21 2.18396
#  _diffrn_orient_matrix_UB_22 -2.60871
#  _diffrn_orient_matrix_UB_23 -9.5302
#  _diffrn_orient_matrix_UB_31 7.4657
#  _diffrn_orient_matrix_UB_32 6.94702
#  _diffrn_orient_matrix_UB_33 -0.18685

#  _phase_label  Fe3O4

#  loop_
#  _diffrn_refln_index_h
#  _diffrn_refln_index_k
#  _diffrn_refln_index_l
#  _diffrn_refln_fr
#  _diffrn_refln_fr_sigma
#  0 0 8 0.64545 0.01329
#  2 0 6 1.75682 0.04540
#  0 2 6 1.67974 0.03711

# """

# obj = Diffrn.from_cif(s_cont)
# print(obj)
# for var_name in obj.get_variable_names():
#     print(var_name)
